"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : May 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.PyQt.QtWidgets import QApplication

from .html_elems import (
    HtmlContent,
    HtmlList,
    HtmlParagraph,
    HtmlSection,
    HtmlTable,
    HtmlTableCol,
    HtmlTableHeader,
)


class DatabaseInfo:
    def __init__(self, db):
        self.db = db

    def __del__(self):
        self.db = None

    def generalInfo(self):
        info = self.db.connector.getInfo()
        tbl = [(QApplication.translate("DBManagerPlugin", "Server version: "), info[0])]
        return HtmlTable(tbl)

    def connectionDetails(self):
        tbl = [
            (
                QApplication.translate("DBManagerPlugin", "Host:"),
                self.db.connector.host,
            ),
            (
                QApplication.translate("DBManagerPlugin", "User:"),
                self.db.connector.user,
            ),
        ]
        return HtmlTable(tbl)

    def spatialInfo(self):
        ret = []

        info = self.db.connector.getSpatialInfo()
        if info is None:
            return

        tbl = [
            (QApplication.translate("DBManagerPlugin", "Library:"), info[0]),
            ("GEOS:", info[1]),
            ("Proj:", info[2]),
        ]
        ret.append(HtmlTable(tbl))

        if not self.db.connector.has_geometry_columns:
            ret.append(
                HtmlParagraph(
                    QApplication.translate(
                        "DBManagerPlugin",
                        "<warning> geometry_columns table doesn't exist!\n"
                        "This table is essential for many GIS applications for enumeration of tables.",
                    )
                )
            )

        return ret

    def privilegesDetails(self):
        details = self.db.connector.getDatabasePrivileges()
        lst = []
        if details[0]:
            lst.append(QApplication.translate("DBManagerPlugin", "create new schemas"))
        if details[1]:
            lst.append(
                QApplication.translate("DBManagerPlugin", "create temporary tables")
            )
        return HtmlList(lst)

    def toHtml(self):
        if self.db is None:
            return HtmlSection(
                QApplication.translate("DBManagerPlugin", "<warning> Not connected")
            ).toHtml()

        ret = []

        # connection details
        conn_details = self.connectionDetails()
        if conn_details is None:
            pass
        else:
            ret.append(
                HtmlSection(
                    QApplication.translate("DBManagerPlugin", "Connection details"),
                    conn_details,
                )
            )

        # database information
        general_info = self.generalInfo()
        if general_info is None:
            pass
        else:
            ret.append(
                HtmlSection(
                    QApplication.translate("DBManagerPlugin", "General info"),
                    general_info,
                )
            )

        # has spatial enabled?
        spatial_info = self.spatialInfo()
        if spatial_info is None:
            pass
        else:
            typename = self.db.connection().typeNameString()
            spatial_info = HtmlContent(spatial_info)
            if not spatial_info.hasContents():
                spatial_info = QApplication.translate(
                    "DBManagerPlugin", "<warning> {0} support not enabled!"
                ).format(typename)
            ret.append(HtmlSection(typename, spatial_info))

        # privileges
        priv_details = self.privilegesDetails()
        if priv_details is None:
            pass
        else:
            priv_details = HtmlContent(priv_details)
            if not priv_details.hasContents():
                priv_details = QApplication.translate(
                    "DBManagerPlugin", "<warning> This user has no privileges!"
                )
            else:
                priv_details = [
                    QApplication.translate("DBManagerPlugin", "User has privileges:"),
                    priv_details,
                ]
            ret.append(
                HtmlSection(
                    QApplication.translate("DBManagerPlugin", "Privileges"),
                    priv_details,
                )
            )

        return HtmlContent(ret).toHtml()


class SchemaInfo:
    def __init__(self, schema):
        self.schema = schema

    def __del__(self):
        self.schema = None

    def generalInfo(self):
        tbl = [
            # ("Tables:", self.schema.tableCount)
        ]
        if self.schema.owner:
            tbl.append(
                (QApplication.translate("DBManagerPlugin", "Owner:"), self.schema.owner)
            )
        if self.schema.comment:
            tbl.append(
                (
                    QApplication.translate("DBManagerPlugin", "Comment:"),
                    self.schema.comment,
                )
            )
        return HtmlTable(tbl)

    def privilegesDetails(self):
        details = self.schema.database().connector.getSchemaPrivileges(self.schema.name)
        lst = []
        if details[0]:
            lst.append(QApplication.translate("DBManagerPlugin", "create new objects"))
        if details[1]:
            lst.append(QApplication.translate("DBManagerPlugin", "access objects"))
        return HtmlList(lst)

    def toHtml(self):
        ret = []

        general_info = self.generalInfo()
        if general_info is None:
            pass
        else:
            ret.append(
                HtmlSection(
                    QApplication.translate("DBManagerPlugin", "Schema details"),
                    general_info,
                )
            )

        priv_details = self.privilegesDetails()
        if priv_details is None:
            pass
        else:
            priv_details = HtmlContent(priv_details)
            if not priv_details.hasContents():
                priv_details = QApplication.translate(
                    "DBManagerPlugin",
                    "<warning> This user has no privileges to access this schema!",
                )
            else:
                priv_details = [
                    QApplication.translate("DBManagerPlugin", "User has privileges:"),
                    priv_details,
                ]
            ret.append(
                HtmlSection(
                    QApplication.translate("DBManagerPlugin", "Privileges"),
                    priv_details,
                )
            )

        return HtmlContent(ret).toHtml()


class TableInfo:
    def __init__(self, table):
        self.table = table

    def __del__(self):
        self.table = None

    def generalInfo(self):
        if self.table.rowCount is None:
            # row count information is not displayed yet, so just block
            # table signals to avoid double refreshing (infoViewer->refreshRowCount->tableChanged->infoViewer)
            self.table.blockSignals(True)
            self.table.refreshRowCount()
            self.table.blockSignals(False)

        tbl = [
            (
                QApplication.translate("DBManagerPlugin", "Relation type:"),
                (
                    QApplication.translate("DBManagerPlugin", "View")
                    if self.table.isView
                    else QApplication.translate("DBManagerPlugin", "Table")
                ),
            ),
            (
                QApplication.translate("DBManagerPlugin", "Rows:"),
                (
                    self.table.rowCount
                    if self.table.rowCount is not None
                    else QApplication.translate(
                        "DBManagerPlugin",
                        'Unknown (<a href="action:rows/count">find out</a>)',
                    )
                ),
            ),
        ]
        if self.table.comment:
            tbl.append(
                (
                    QApplication.translate("DBManagerPlugin", "Comment:"),
                    self.table.comment,
                )
            )

        return HtmlTable(tbl)

    def spatialInfo(self):  # implemented in subclasses
        return None

    def fieldsDetails(self):
        tbl = []

        # define the table header
        header = (
            "#",
            QApplication.translate("DBManagerPlugin", "Name"),
            QApplication.translate("DBManagerPlugin", "Type"),
            QApplication.translate("DBManagerPlugin", "Null"),
            QApplication.translate("DBManagerPlugin", "Default"),
        )
        tbl.append(HtmlTableHeader(header))

        # add table contents
        for fld in self.table.fields():
            is_null_txt = "N" if fld.notNull else "Y"

            # make primary key field underlined
            attrs = {"class": "underline"} if fld.primaryKey else None
            name = HtmlTableCol(fld.name, attrs)

            tbl.append(
                (fld.num, name, fld.type2String(), is_null_txt, fld.default2String())
            )

        return HtmlTable(tbl, {"class": "header"})

    def constraintsDetails(self):
        if self.table.constraints() is None or len(self.table.constraints()) <= 0:
            return None

        tbl = []

        # define the table header
        header = (
            QApplication.translate("DBManagerPlugin", "Name"),
            QApplication.translate("DBManagerPlugin", "Type"),
            QApplication.translate("DBManagerPlugin", "Column(s)"),
        )
        tbl.append(HtmlTableHeader(header))

        # add table contents
        for con in self.table.constraints():
            # get the fields the constraint is defined on
            cols = [
                p[1].name if p[1] is not None else "??? (#%d)" % p[0]
                for p in iter(list(con.fields().items()))
            ]
            tbl.append((con.name, con.type2String(), "\n".join(cols)))

        return HtmlTable(tbl, {"class": "header"})

    def indexesDetails(self):
        if self.table.indexes() is None or len(self.table.indexes()) <= 0:
            return None

        tbl = []

        # define the table header
        header = (
            QApplication.translate("DBManagerPlugin", "Name"),
            QApplication.translate("DBManagerPlugin", "Column(s)"),
        )
        tbl.append(HtmlTableHeader(header))

        # add table contents
        for idx in self.table.indexes():
            # get the fields the index is defined on
            cols = [
                p[1].name if p[1] is not None else "??? (#%d)" % p[0]
                for p in iter(list(idx.fields().items()))
            ]
            tbl.append((idx.name, "\n".join(cols)))

        return HtmlTable(tbl, {"class": "header"})

    def triggersDetails(self):
        if self.table.triggers() is None or len(self.table.triggers()) <= 0:
            return None

        tbl = []

        # define the table header
        header = (
            QApplication.translate("DBManagerPlugin", "Name"),
            QApplication.translate("DBManagerPlugin", "Function"),
        )
        tbl.append(HtmlTableHeader(header))

        # add table contents
        for trig in self.table.triggers():
            name = (
                '{name} (<a href="action:trigger/{name}/{action}">{action}</a>)'.format(
                    name=trig.name, action="delete"
                )
            )
            tbl.append((name, trig.function.replace("<", "&lt;")))

        return HtmlTable(tbl, {"class": "header"})

    def getViewDefinition(self):
        if not self.table.isView:
            return None
        return self.table.database().connector.getViewDefinition(
            (self.table.schemaName(), self.table.name)
        )

    def getTableInfo(self):
        ret = []

        general_info = self.generalInfo()
        if general_info is None:
            pass
        else:
            ret.append(
                HtmlSection(
                    QApplication.translate("DBManagerPlugin", "General info"),
                    general_info,
                )
            )

        # spatial info
        spatial_info = self.spatialInfo()
        if spatial_info is None:
            pass
        else:
            spatial_info = HtmlContent(spatial_info)
            if not spatial_info.hasContents():
                spatial_info = QApplication.translate(
                    "DBManagerPlugin", "<warning> This is not a spatial table."
                )
            ret.append(
                HtmlSection(
                    self.table.database().connection().typeNameString(), spatial_info
                )
            )

        # fields
        fields_details = self.fieldsDetails()
        if fields_details is None:
            pass
        else:
            ret.append(
                HtmlSection(
                    QApplication.translate("DBManagerPlugin", "Fields"), fields_details
                )
            )

        # constraints
        constraints_details = self.constraintsDetails()
        if constraints_details is None:
            pass
        else:
            ret.append(
                HtmlSection(
                    QApplication.translate("DBManagerPlugin", "Constraints"),
                    constraints_details,
                )
            )

        # indexes
        indexes_details = self.indexesDetails()
        if indexes_details is None:
            pass
        else:
            ret.append(
                HtmlSection(
                    QApplication.translate("DBManagerPlugin", "Indexes"),
                    indexes_details,
                )
            )

        # triggers
        triggers_details = self.triggersDetails()
        if triggers_details is None:
            pass
        else:
            ret.append(
                HtmlSection(
                    QApplication.translate("DBManagerPlugin", "Triggers"),
                    triggers_details,
                )
            )

        return ret

    def getViewInfo(self):
        if not self.table.isView:
            return []

        ret = self.getTableInfo()

        # view definition
        view_def = self.getViewDefinition()
        if view_def is None:
            pass
        else:
            ret.append(
                HtmlSection(
                    QApplication.translate("DBManagerPlugin", "View definition"),
                    view_def,
                )
            )

        return ret

    def toHtml(self):
        if self.table.isView:
            ret = self.getViewInfo()
        else:
            ret = self.getTableInfo()
        return HtmlContent(ret).toHtml()


class VectorTableInfo(TableInfo):
    def __init__(self, table):
        TableInfo.__init__(self, table)

    def spatialInfo(self):
        ret = []
        if self.table.geomType is None:
            return ret

        tbl = [
            (
                QApplication.translate("DBManagerPlugin", "Column:"),
                self.table.geomColumn,
            ),
            (
                QApplication.translate("DBManagerPlugin", "Geometry:"),
                self.table.geomType,
            ),
        ]

        # only if we have info from geometry_columns
        if self.table.geomDim:
            tbl.append(
                (
                    QApplication.translate("DBManagerPlugin", "Dimension:"),
                    self.table.geomDim,
                )
            )

        srid = self.table.srid if self.table.srid not in (None, 0) else -1
        sr_info = (
            self.table.database().connector.getSpatialRefInfo(srid)
            if srid != -1
            else QApplication.translate("DBManagerPlugin", "Undefined")
        )
        if sr_info:
            tbl.append(
                (
                    QApplication.translate("DBManagerPlugin", "Spatial ref:"),
                    "%s (%d)" % (sr_info, srid),
                )
            )

        # estimated extent
        if not self.table.isView:
            if self.table.estimatedExtent is None:
                # estimated extent information is not displayed yet, so just block
                # table signals to avoid double refreshing (infoViewer->refreshEstimatedExtent->tableChanged->infoViewer)
                self.table.blockSignals(True)
                self.table.refreshTableEstimatedExtent()
                self.table.blockSignals(False)

            if (
                self.table.estimatedExtent is not None
                and self.table.estimatedExtent[0] is not None
            ):
                if isinstance(self.table.estimatedExtent, list):
                    estimated_extent_str = ", ".join(
                        "%.5f" % e for e in self.table.estimatedExtent
                    )
                else:
                    estimated_extent_str = (
                        "%.5f, %.5f - %.5f, %.5f" % self.table.estimatedExtent
                    )
                tbl.append(
                    (
                        QApplication.translate("DBManagerPlugin", "Estimated extent:"),
                        estimated_extent_str,
                    )
                )

        # extent
        if self.table.extent is not None and self.table.extent[0] is not None:
            if isinstance(self.table.extent, list):
                extent_str = ", ".join("%.5f" % e for e in self.table.extent)
            else:
                extent_str = "%.5f, %.5f - %.5f, %.5f" % self.table.extent
        else:
            extent_str = QApplication.translate(
                "DBManagerPlugin",
                '(unknown) (<a href="action:extent/get">find out</a>)',
            )
        tbl.append((QApplication.translate("DBManagerPlugin", "Extent:"), extent_str))

        ret.append(HtmlTable(tbl))

        # is there an entry in geometry_columns?
        if self.table.geomType.lower() == "geometry":
            ret.append(
                HtmlParagraph(
                    QApplication.translate(
                        "DBManagerPlugin",
                        "<warning> There is no entry in geometry_columns!",
                    )
                )
            )

        # find out whether the geometry column has spatial index on it
        if not self.table.isView:
            if not self.table.hasSpatialIndex():
                ret.append(
                    HtmlParagraph(
                        QApplication.translate(
                            "DBManagerPlugin",
                            '<warning> No spatial index defined (<a href="action:spatialindex/create">create it</a>)',
                        )
                    )
                )

        return ret


class RasterTableInfo(TableInfo):
    def __init__(self, table):
        TableInfo.__init__(self, table)

    def spatialInfo(self):
        ret = []
        if self.table.geomType is None:
            return ret

        tbl = [
            (
                QApplication.translate("DBManagerPlugin", "Column:"),
                self.table.geomColumn,
            ),
            (
                QApplication.translate("DBManagerPlugin", "Geometry:"),
                self.table.geomType,
            ),
        ]

        # only if we have info from geometry_columns
        srid = self.table.srid if self.table.srid is not None else -1
        sr_info = (
            self.table.database().connector.getSpatialRefInfo(srid)
            if srid != -1
            else QApplication.translate("DBManagerPlugin", "Undefined")
        )
        if sr_info:
            tbl.append(
                (
                    QApplication.translate("DBManagerPlugin", "Spatial ref:"),
                    "%s (%d)" % (sr_info, srid),
                )
            )

        # extent
        if self.table.extent is not None and self.table.extent[0] is not None:
            extent_str = "%.5f, %.5f - %.5f, %.5f" % self.table.extent
        else:
            extent_str = QApplication.translate(
                "DBManagerPlugin",
                '(unknown) (<a href="action:extent/get">find out</a>)',
            )
        tbl.append((QApplication.translate("DBManagerPlugin", "Extent:"), extent_str))

        ret.append(HtmlTable(tbl))
        return ret



"""
***************************************************************************
    HypsometricCurves.py
    ---------------------
    Date                 : November 2014
    Copyright            : (C) 2014 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "November 2014"
__copyright__ = "(C) 2014, Alexander Bruy"

import csv
import os

from osgeo import gdal, ogr, osr

gdal.UseExceptions()
ogr.UseExceptions()
osr.UseExceptions()

from qgis.core import (
    QgsFeatureRequest,
    QgsGeometry,
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterFolderDestination,
    QgsProcessingParameterNumber,
    QgsProcessingParameterRasterLayer,
    QgsRectangle,
)
from qgis.PyQt.QtCore import QCoreApplication

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm
from processing.tools import raster


class HypsometricCurves(QgisAlgorithm):
    INPUT_DEM = "INPUT_DEM"
    BOUNDARY_LAYER = "BOUNDARY_LAYER"
    STEP = "STEP"
    USE_PERCENTAGE = "USE_PERCENTAGE"
    OUTPUT_DIRECTORY = "OUTPUT_DIRECTORY"

    def group(self):
        return self.tr("Raster terrain analysis")

    def groupId(self):
        return "rasterterrainanalysis"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterRasterLayer(self.INPUT_DEM, self.tr("DEM to analyze"))
        )
        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.BOUNDARY_LAYER,
                self.tr("Boundary layer"),
                [QgsProcessing.SourceType.TypeVectorPolygon],
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.STEP,
                self.tr("Step"),
                type=QgsProcessingParameterNumber.Type.Double,
                minValue=0.0,
                defaultValue=100.0,
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.USE_PERCENTAGE,
                self.tr("Use % of area instead of absolute value"),
                defaultValue=False,
            )
        )

        self.addParameter(
            QgsProcessingParameterFolderDestination(
                self.OUTPUT_DIRECTORY, self.tr("Hypsometric curves")
            )
        )

    def name(self):
        return "hypsometriccurves"

    def displayName(self):
        return self.tr("Hypsometric curves")

    def shortDescription(self):
        return self.tr(
            "Computes hypsometric curves for an input Digital Elevation Model (DEM) as table files."
        )

    def shortHelpString(self):
        return self.tr(
            "This algorithm computes hypsometric curves for an input Digital Elevation Model (DEM). "
            "Curves are produced as table files in an output folder specified by the user."
        )

    def processAlgorithm(self, parameters, context, feedback):
        try:
            import numpy
        except ImportError:
            raise QgsProcessingException(
                QCoreApplication.translate(
                    "HypsometricCurves",
                    "This algorithm requires the Python “numpy” library. Please install this library and try again.",
                )
            )

        raster_layer = self.parameterAsRasterLayer(parameters, self.INPUT_DEM, context)
        target_crs = raster_layer.crs()
        rasterPath = raster_layer.source()

        source = self.parameterAsSource(parameters, self.BOUNDARY_LAYER, context)
        if source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.BOUNDARY_LAYER)
            )

        step = self.parameterAsDouble(parameters, self.STEP, context)
        percentage = self.parameterAsBoolean(parameters, self.USE_PERCENTAGE, context)

        outputPath = self.parameterAsString(parameters, self.OUTPUT_DIRECTORY, context)
        if not os.path.exists(outputPath):
            os.makedirs(outputPath)

        rasterDS = gdal.Open(rasterPath, gdal.GA_ReadOnly)
        geoTransform = rasterDS.GetGeoTransform()
        rasterBand = rasterDS.GetRasterBand(1)
        noData = rasterBand.GetNoDataValue()

        cellXSize = abs(geoTransform[1])
        cellYSize = abs(geoTransform[5])
        rasterXSize = rasterDS.RasterXSize
        rasterYSize = rasterDS.RasterYSize

        rasterBBox = QgsRectangle(
            geoTransform[0],
            geoTransform[3] - cellYSize * rasterYSize,
            geoTransform[0] + cellXSize * rasterXSize,
            geoTransform[3],
        )
        rasterGeom = QgsGeometry.fromRect(rasterBBox)

        crs = osr.SpatialReference()
        crs.ImportFromProj4(str(target_crs.toProj()))

        memVectorDriver = ogr.GetDriverByName("Memory")
        memRasterDriver = gdal.GetDriverByName("MEM")

        features = source.getFeatures(
            QgsFeatureRequest().setDestinationCrs(
                target_crs, context.transformContext()
            )
        )
        total = 100.0 / source.featureCount() if source.featureCount() else 0

        for current, f in enumerate(features):
            if not f.hasGeometry():
                continue

            if feedback.isCanceled():
                break

            geom = f.geometry()
            intersectedGeom = rasterGeom.intersection(geom)

            if intersectedGeom.isEmpty():
                feedback.pushInfo(
                    self.tr(
                        "Feature {0} does not intersect raster or "
                        "entirely located in NODATA area"
                    ).format(f.id())
                )
                continue

            fName = os.path.join(
                outputPath, f"histogram_{source.sourceName()}_{f.id()}.csv"
            )

            ogrGeom = ogr.CreateGeometryFromWkt(intersectedGeom.asWkt())
            bbox = intersectedGeom.boundingBox()
            xMin = bbox.xMinimum()
            xMax = bbox.xMaximum()
            yMin = bbox.yMinimum()
            yMax = bbox.yMaximum()

            (startColumn, startRow) = raster.mapToPixel(xMin, yMax, geoTransform)
            (endColumn, endRow) = raster.mapToPixel(xMax, yMin, geoTransform)

            width = endColumn - startColumn
            height = endRow - startRow

            srcOffset = (startColumn, startRow, width, height)
            srcArray = rasterBand.ReadAsArray(*srcOffset)

            if srcOffset[2] == 0 or srcOffset[3] == 0:
                feedback.pushInfo(
                    self.tr("Feature {0} is smaller than raster cell size").format(
                        f.id()
                    )
                )
                continue

            newGeoTransform = (
                geoTransform[0] + srcOffset[0] * geoTransform[1],
                geoTransform[1],
                0.0,
                geoTransform[3] + srcOffset[1] * geoTransform[5],
                0.0,
                geoTransform[5],
            )

            memVDS = memVectorDriver.CreateDataSource("out")
            memLayer = memVDS.CreateLayer("poly", crs, ogr.wkbPolygon)

            ft = ogr.Feature(memLayer.GetLayerDefn())
            ft.SetGeometry(ogrGeom)
            memLayer.CreateFeature(ft)
            ft.Destroy()

            rasterizedDS = memRasterDriver.Create(
                "", srcOffset[2], srcOffset[3], 1, gdal.GDT_Byte
            )
            rasterizedDS.SetGeoTransform(newGeoTransform)
            gdal.RasterizeLayer(rasterizedDS, [1], memLayer, burn_values=[1])
            rasterizedArray = rasterizedDS.ReadAsArray()

            srcArray = numpy.nan_to_num(srcArray)
            masked = numpy.ma.MaskedArray(
                srcArray,
                mask=numpy.logical_or(
                    srcArray == noData, numpy.logical_not(rasterizedArray)
                ),
            )

            self.calculateHypsometry(
                f.id(), fName, feedback, masked, cellXSize, cellYSize, percentage, step
            )

            memVDS = None
            rasterizedDS = None
            feedback.setProgress(int(current * total))

        rasterDS = None

        return {self.OUTPUT_DIRECTORY: outputPath}

    def calculateHypsometry(self, fid, fName, feedback, data, pX, pY, percentage, step):
        out = dict()
        d = data.compressed()
        if d.size == 0:
            feedback.pushInfo(
                self.tr(
                    "Feature {0} does not intersect raster or "
                    "entirely located in NODATA area"
                ).format(fid)
            )
            return

        minValue = d.min()
        maxValue = d.max()
        startValue = minValue
        tmpValue = minValue + step
        while startValue < maxValue:
            out[tmpValue] = ((startValue <= d) & (d < tmpValue)).sum()
            startValue = tmpValue
            tmpValue += step

        if percentage:
            multiplier = 100.0 / len(d.flat)
        else:
            multiplier = pX * pY

        for k, v in out.items():
            out[k] = v * multiplier

        prev = None
        for i in sorted(out.items()):
            if prev is None:
                out[i[0]] = i[1]
            else:
                out[i[0]] = i[1] + out[prev]
            prev = i[0]

        with open(fName, "w", newline="", encoding="utf-8") as out_file:
            writer = csv.writer(out_file)
            writer.writerow([self.tr("Area"), self.tr("Elevation")])

            for i in sorted(out.items()):
                writer.writerow([i[1], i[0]])



"""
***************************************************************************
    ModelConfigWidgets.py
    ---------------------
    Date                 : January 2026
    Copyright            : (C) 2016 by Nyall Dawson
    Email                : nyall dot dawson at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

from qgis.core import (
    QgsProcessingContext,
    QgsProcessingModelChildAlgorithm,
    QgsProcessingModelOutput,
    QgsProcessingModelParameter,
    QgsProcessingParameterDefinition,
)
from qgis.gui import (
    QgsProcessingModelConfigWidgetFactory,
    QgsProcessingParameterDefinitionPanelWidget,
    QgsProcessingParameterWidgetContext,
)

from processing.modeler.ModelerParameterDefinitionDialog import (
    ModelerParameterDefinitionDialog,
    ModelerParameterDefinitionWidget,
)
from processing.modeler.ModelerParametersDialog import (
    ModelerParametersDialog,
    ModelerParametersWidget,
)


class ModelConfigWidgetFactory(QgsProcessingModelConfigWidgetFactory):
    def supportsComponent(self, component):
        return isinstance(
            component,
            (
                QgsProcessingModelParameter,
                QgsProcessingModelOutput,
                QgsProcessingModelChildAlgorithm,
            ),
        )

    def createWidget(
        self,
        component,
        context: QgsProcessingContext,
        widgetContext: QgsProcessingParameterWidgetContext,
    ):
        model = widgetContext.model()
        if not model:
            return None

        model_dialog = widgetContext.modelDesignerDialog()

        if isinstance(component, QgsProcessingModelParameter):
            component_name = component.parameterName()
            existing_param = model.parameterDefinition(component_name)
            if ModelerParameterDefinitionDialog.use_legacy_dialog(param=existing_param):
                # boo, old api, not supported for the dock
                return None

            comment = component.comment().description()
            comment_color = component.comment().color()
            old_name = existing_param.name()
            old_description = existing_param.description()

            widget = QgsProcessingParameterDefinitionPanelWidget(
                type=existing_param.type(),
                context=context,
                widgetContext=widgetContext,
                definition=existing_param,
                algorithm=model,
            )
            widget.setComments(comment)
            widget.setCommentColor(comment_color)
            if widgetContext.processingContextGenerator():
                widget.registerProcessingContextGenerator(
                    widgetContext.processingContextGenerator()
                )

            existing_param_name = existing_param.name()

            def on_widget_changed():
                nonlocal existing_param_name
                model_scene = model_dialog.modelScene()
                graphic_item = model_scene.parameterItem(component_name)
                if not graphic_item:
                    # should not happen!
                    return

                new_param = widget.createParameter(existing_param_name)
                comment = widget.comments()
                comment_color = widget.commentColor()
                existing_param_name = graphic_item.apply_new_param(
                    new_param, old_description, old_name, comment, comment_color
                )

            widget.widgetChanged.connect(on_widget_changed)
            return widget
        elif isinstance(component, QgsProcessingModelOutput):
            child_id = component.childId()
            child_output_name = component.childOutputName()

            child_alg = model.childAlgorithm(child_id)
            comment = component.comment().description()
            comment_color = component.comment().color()

            existing_param = model.modelParameterFromChildIdAndOutputName(
                component.childId(), component.name()
            )

            widget = ModelerParameterDefinitionWidget(
                model,
                param=existing_param,
            )
            widget.setComments(comment)
            widget.setCommentColor(comment_color)

            def on_widget_changed():
                model_scene = model_dialog.modelScene()
                graphic_item = model_scene.outputItem(child_id, child_output_name)
                if not graphic_item:
                    # should not happen!
                    return

                new_param = widget.create_parameter()
                graphic_item.apply_new_output(
                    name=new_param.description(),
                    description=new_param.description(),
                    default=new_param.defaultValue(),
                    mandatory=not (
                        new_param.flags()
                        & QgsProcessingParameterDefinition.Flag.FlagOptional
                    ),
                    comments=widget.comments(),
                    comment_color=widget.commentColor(),
                    child_alg=child_alg,
                )

            widget.widgetChanged.connect(on_widget_changed)
            return widget
        elif isinstance(component, QgsProcessingModelChildAlgorithm):
            algorithm = component.algorithm().create()
            context = widgetContext.processingContextGenerator().processingContext()
            child_id = component.childId()

            # this is unfortunately VERY ugly. The old Python processing GUI api
            # had the ModelerParametersDialog class so enmeshed in the parameter
            # GUI classes that there is no way to rip this out without breaking
            # API. So we HAVE to do a gross hack here and make a dialog, which
            # we never show, but which allows us to keep the old API intact...
            fake_dialog = ModelerParametersDialog(
                algorithm,
                model,
                algName=child_id,
                configuration=component.configuration(),
            )

            widget = ModelerParametersWidget(
                algorithm,
                model,
                algName=child_id,
                configuration=component.configuration(),
                dialog=fake_dialog,
                context=context,
            )
            widget.setComments(component.comment().description())
            widget.setCommentColor(component.comment().color())

            # make sure fake dialog exists for lifetime of widget
            widget.__fake_dialog = fake_dialog

            def on_widget_changed():
                model_scene = model_dialog.modelScene()
                graphic_item = model_scene.childAlgorithmItem(child_id)
                if not graphic_item:
                    # should not happen!
                    return

                graphic_item.apply_new_alg(widget.createAlgorithm())

            widget.widgetChanged.connect(on_widget_changed)
            return widget

        return None



"""
***************************************************************************
    v_transform.py
    ----------
    Date                 : February 2024
    Copyright            : (C) 2024 by Andrea Giudiceandrea
    Email                : andreaerdna at libero dot it
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   any later version.                                                    *
*                                                                         *
***************************************************************************
"""

__author__ = "Andrea Giudiceandrea"
__date__ = "February 2024"
__copyright__ = "(C) 2024, Andrea Giudiceandrea"

from qgis.PyQt.QtCore import QCoreApplication


def checkParameterValuesBeforeExecuting(alg, parameters, context):
    """Verify if we have the right parameters"""
    # -w, -x and -y parameters are mutually exclusive
    w = alg.parameterAsBoolean(parameters, "-w", context)
    x = alg.parameterAsBoolean(parameters, "-x", context)
    y = alg.parameterAsBoolean(parameters, "-y", context)
    if sum([w, x, y]) > 1:
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt",
            "The 'Swap coordinates' parameters -w, -x and -y are mutually exclusive. You need to set either none or only one of them!",
        )

    return True, None



"""
***************************************************************************
    RasterLayerCalculator.py
    ---------------------
    Date                 : November 2016
    Copyright            : (C) 2016 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "November 2016"
__copyright__ = "(C) 2016, Victor Olaya"

import math
import os

from qgis.analysis import QgsRasterCalculator, QgsRasterCalculatorEntry
from qgis.core import (
    QgsCoordinateTransform,
    QgsMapLayer,
    QgsProcessing,
    QgsProcessingAlgorithm,
    QgsProcessingException,
    QgsProcessingOutputRasterLayer,
    QgsProcessingParameterCrs,
    QgsProcessingParameterExtent,
    QgsProcessingParameterMultipleLayers,
    QgsProcessingParameterNumber,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterString,
    QgsProcessingUtils,
)
from qgis.PyQt.QtCore import QObject

from processing.algs.gdal.GdalUtils import GdalUtils
from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm


class RasterCalculator(QgisAlgorithm):
    LAYERS = "LAYERS"
    EXTENT = "EXTENT"
    CELLSIZE = "CELLSIZE"
    EXPRESSION = "EXPRESSION"
    CRS = "CRS"
    OUTPUT = "OUTPUT"

    def group(self):
        return self.tr("Raster analysis")

    def groupId(self):
        return "rasteranalysis"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        class ParameterRasterCalculatorExpression(QgsProcessingParameterString):
            def __init__(self, name="", description="", multiLine=False):
                super().__init__(name, description, multiLine=multiLine)
                self.setMetadata(
                    {
                        "widget_wrapper": "processing.algs.qgis.ui.RasterCalculatorWidgets.ExpressionWidgetWrapper"
                    }
                )

            def type(self):
                return "raster_calc_expression"

            def clone(self):
                return ParameterRasterCalculatorExpression(
                    self.name(), self.description(), self.multiLine()
                )

        self.addParameter(
            ParameterRasterCalculatorExpression(
                self.EXPRESSION, self.tr("Expression"), multiLine=True
            )
        )
        self.addParameter(
            QgsProcessingParameterMultipleLayers(
                self.LAYERS,
                self.tr(
                    "Reference layer(s) (used for automated extent, cellsize, and CRS)"
                ),
                layerType=QgsProcessing.SourceType.TypeRaster,
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.CELLSIZE,
                self.tr("Cell size (use 0 or empty to set it automatically)"),
                type=QgsProcessingParameterNumber.Type.Double,
                minValue=0.0,
                defaultValue=0.0,
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterExtent(
                self.EXTENT, self.tr("Output extent"), optional=True
            )
        )
        self.addParameter(
            QgsProcessingParameterCrs(self.CRS, "Output CRS", optional=True)
        )
        self.addParameter(
            QgsProcessingParameterRasterDestination(self.OUTPUT, self.tr("Output"))
        )

    def flags(self):
        return super().flags() | QgsProcessingAlgorithm.Flag.FlagDeprecated

    def name(self):
        return "rastercalculator"

    def displayName(self):
        return self.tr("Raster calculator")

    def shortDescription(self):
        return self.tr("Allows performing algebraic operations using raster layers.")

    def shortHelpString(self):
        return self.tr(
            "This algorithm allows performing algebraic operations using raster layers. \n"
            "The resulting layer will have its values computed according to an expression. "
            "The expression can contain numerical values, operators and references to any of "
            "the layers in the current project. The following functions are also supported:\n"
            "- sin(), cos(), tan(), atan2(), ln(), log10()\n"
            "The extent, cell size, and output CRS can be defined by the user. "
            "If the extent is not specified, the minimum extent that covers selected reference "
            "layer(s) will be used. If the cell size is not specified, the minimum cell size "
            "of selected reference layer(s) will be used. If the output CRS is not specified, "
            "the CRS of the first reference layer will be used.\n"
            "The cell size is assumed to be the same in both X and Y axes.\n"
            "Layers are referred by their name as displayed in the layer list "
            "and the number of the band to use (based on 1), using the pattern 'layer_name@band number'. "
            "For instance, the first band from a layer named DEM will be referred as DEM@1.\n"
            "When using the calculator in the batch interface or from the console, the files to use have to be specified. "
            "The corresponding layers are referred using the base name of the file (without the full path). "
            "For instance, if using a layer at path/to/my/rasterfile.tif, "
            "the first band of that layer will be referred as rasterfile.tif@1."
        )

    def processAlgorithm(self, parameters, context, feedback):
        expression = self.parameterAsString(parameters, self.EXPRESSION, context)
        layers = self.parameterAsLayerList(parameters, self.LAYERS, context)

        layersDict = {}
        if layers:
            layersDict = {lyr.source(): lyr for lyr in layers}

        crs = self.parameterAsCrs(parameters, self.CRS, context)
        if crs is None or not crs.isValid():
            if not layers:
                raise QgsProcessingException(
                    self.tr("No reference layer selected nor CRS provided")
                )
            else:
                crs = list(layersDict.values())[0].crs()

        bbox = self.parameterAsExtent(parameters, self.EXTENT, context)
        if bbox.isNull() and not layers:
            raise QgsProcessingException(
                self.tr("No reference layer selected nor extent box provided")
            )

        if not bbox.isNull():
            bboxCrs = self.parameterAsExtentCrs(parameters, self.EXTENT, context)
            if bboxCrs != crs:
                transform = QgsCoordinateTransform(
                    bboxCrs, crs, context.transformContext()
                )
                bbox = transform.transformBoundingBox(bbox)

        if bbox.isNull() and layers:
            bbox = QgsProcessingUtils.combineLayerExtents(layers, crs, context)

        cellsize = self.parameterAsDouble(parameters, self.CELLSIZE, context)
        if cellsize == 0 and not layers:
            raise QgsProcessingException(
                self.tr("No reference layer selected nor cellsize value provided")
            )

        def _cellsize(layer):
            ext = layer.extent()
            if layer.crs() != crs:
                transform = QgsCoordinateTransform(
                    layer.crs(), crs, context.transformContext()
                )
                ext = transform.transformBoundingBox(ext)
            return (ext.xMaximum() - ext.xMinimum()) / layer.width()

        if cellsize == 0:
            cellsize = min([_cellsize(lyr) for lyr in layersDict.values()])

        # check for layers available in the model
        layersDictCopy = (
            layersDict.copy()
        )  # need a shallow copy because next calls invalidate iterator
        for lyr in layersDictCopy.values():
            expression = self.mappedNameToLayer(lyr, expression, layersDict, context)

        # check for layers available in the project
        if context.project():
            for lyr in QgsProcessingUtils.compatibleRasterLayers(context.project()):
                expression = self.mappedNameToLayer(
                    lyr, expression, layersDict, context
                )

        # create the list of layers to be passed as inputs to RasterCalculaltor
        # at this phase expression has been modified to match available layers
        # in the current scope
        entries = []
        for name, lyr in layersDict.items():
            for n in range(lyr.bandCount()):
                ref = f"{name:s}@{n + 1:d}"

                if ref in expression:
                    entry = QgsRasterCalculatorEntry()
                    entry.ref = ref
                    entry.raster = lyr
                    entry.bandNumber = n + 1
                    entries.append(entry)

        # Append any missing entry from the current project
        for entry in QgsRasterCalculatorEntry.rasterEntries():
            if not [e for e in entries if e.ref == entry.ref]:
                entries.append(entry)

        output = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)

        width = round((bbox.xMaximum() - bbox.xMinimum()) / cellsize)
        height = round((bbox.yMaximum() - bbox.yMinimum()) / cellsize)
        driverName = GdalUtils.getFormatShortNameFromFilename(output)

        calc = QgsRasterCalculator(
            expression,
            output,
            driverName,
            bbox,
            crs,
            width,
            height,
            entries,
            context.transformContext(),
        )

        res = calc.processCalculation(feedback)
        if res == QgsRasterCalculator.Result.ParserError:
            raise QgsProcessingException(self.tr("Error parsing formula"))
        elif res == QgsRasterCalculator.Result.CalculationError:
            raise QgsProcessingException(
                self.tr("An error occurred while performing the calculation")
            )

        return {self.OUTPUT: output}

    def mappedNameToLayer(self, lyr, expression, layersDict, context):
        """Try to identify if a real layer is mapped in the expression with a symbolic name."""

        nameToMap = lyr.source()

        # check if nameToMap is a file
        # TODO: what about URI eg for a COG?
        if os.path.isfile(nameToMap):
            # get only the name without extension and path of the file
            nameToMap = os.path.splitext(os.path.basename(nameToMap))[0]

        # check for layers directly added in the expression
        if (nameToMap + "@") in expression:
            layersDict[nameToMap] = lyr

        # get "algorithm_inputs" scope of the expressionContext related
        # with mapped variables
        indexOfScope = context.expressionContext().indexOfScope("algorithm_inputs")
        if indexOfScope >= 0:
            expContextAlgInputsScope = context.expressionContext().scope(indexOfScope)

            # check for the layers that are mapped as input in a model
            # to do this check in the latest scope all passed variables
            # to look for a variable that is a layer or a string filename
            # to a layer
            varDescription = None
            for varName in expContextAlgInputsScope.variableNames():
                layerInContext = expContextAlgInputsScope.variable(varName)

                if not isinstance(layerInContext, str) and not isinstance(
                    layerInContext, QgsMapLayer
                ):
                    continue

                if (
                    isinstance(layerInContext, QgsMapLayer)
                    and nameToMap not in layerInContext.source()
                ):
                    continue

                varDescription = expContextAlgInputsScope.description(varName)

                # because there can be variable with None or "" description
                # then skip them
                if not varDescription:
                    continue

                # check if it's description starts with Output as in:
                #    Output 'Output' from algorithm 'calc1'
                # as set in https://github.com/qgis/QGIS/blob/master/src/core/processing/models/qgsprocessingmodelalgorithm.cpp#L516
                # but var in expression is called simply
                #    'Output' from algorithm 'calc1'

                # get the translation string to use to parse the description
                # HAVE to use the same translated string as in
                # https://github.com/qgis/QGIS/blob/master/src/core/processing/models/qgsprocessingmodelalgorithm.cpp#L516
                translatedDesc = self.tr("Output '%1' from algorithm '%2'")
                elementZero = translatedDesc.split(" ")[
                    0
                ]  # For english the string result should be "Output"

                elements = varDescription.split(" ")
                if len(elements) > 1 and elements[0] == elementZero:
                    # remove heading QObject.tr"Output ") string. Note adding a space at the end of elementZero!
                    varDescription = varDescription[len(elementZero) + 1 :]

                # check if cleaned varDescription is present in the expression
                # if not skip it
                if (varDescription + "@") not in expression:
                    continue

                # !!!found!!! => substitute in expression
                # and add in the list of layers that will be passed to raster calculator
                nameToMap = varName
                new = f"{nameToMap}@"
                old = f"{varDescription}@"
                expression = expression.replace(old, new)

                layersDict[nameToMap] = lyr

        # need return the modified expression because it's not a reference
        return expression



"""
***************************************************************************
    Dissolve.py
    ---------------------
    Date                 : Janaury 2015
    Copyright            : (C) 2015 by Giovanni Manghi
    Email                : giovanni dot manghi at naturalgis dot pt
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Giovanni Manghi"
__date__ = "January 2015"
__copyright__ = "(C) 2015, Giovanni Manghi"

from qgis.core import (
    QgsProcessingException,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingParameterString,
    QgsProcessingParameterVectorDestination,
)

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils


class Dissolve(GdalAlgorithm):
    INPUT = "INPUT"
    FIELD = "FIELD"
    GEOMETRY = "GEOMETRY"
    EXPLODE_COLLECTIONS = "EXPLODE_COLLECTIONS"
    KEEP_ATTRIBUTES = "KEEP_ATTRIBUTES"
    COUNT_FEATURES = "COUNT_FEATURES"
    COMPUTE_AREA = "COMPUTE_AREA"
    COMPUTE_STATISTICS = "COMPUTE_STATISTICS"
    STATISTICS_ATTRIBUTE = "STATISTICS_ATTRIBUTE"
    OPTIONS = "OPTIONS"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterFeatureSource(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.FIELD,
                self.tr("Dissolve field"),
                None,
                self.INPUT,
                QgsProcessingParameterField.DataType.Any,
                optional=True,
            )
        )
        self.addParameter(
            QgsProcessingParameterString(
                self.GEOMETRY, self.tr("Geometry column name"), defaultValue="geometry"
            )
        )
        params = [
            QgsProcessingParameterBoolean(
                self.EXPLODE_COLLECTIONS,
                self.tr(
                    "Produce one feature for each geometry in any kind of geometry collection in the source file"
                ),
                defaultValue=False,
            ),
            QgsProcessingParameterBoolean(
                self.KEEP_ATTRIBUTES,
                self.tr("Keep input attributes"),
                defaultValue=False,
            ),
            QgsProcessingParameterBoolean(
                self.COUNT_FEATURES,
                self.tr("Count dissolved features"),
                defaultValue=False,
            ),
            QgsProcessingParameterBoolean(
                self.COMPUTE_AREA,
                self.tr("Compute area and perimeter of dissolved features"),
                defaultValue=False,
            ),
            QgsProcessingParameterBoolean(
                self.COMPUTE_STATISTICS,
                self.tr("Compute min/max/sum/mean for attribute"),
                defaultValue=False,
            ),
            QgsProcessingParameterField(
                self.STATISTICS_ATTRIBUTE,
                self.tr("Numeric attribute to calculate statistics on"),
                None,
                self.INPUT,
                QgsProcessingParameterField.DataType.Numeric,
                optional=True,
            ),
            QgsProcessingParameterString(
                self.OPTIONS,
                self.tr("Additional creation options"),
                defaultValue="",
                optional=True,
            ),
        ]
        for param in params:
            param.setFlags(
                param.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced
            )
            self.addParameter(param)

        self.addParameter(
            QgsProcessingParameterVectorDestination(self.OUTPUT, self.tr("Dissolved"))
        )

    def name(self):
        return "dissolve"

    def displayName(self):
        return self.tr("Dissolve")

    def group(self):
        return self.tr("Vector geoprocessing")

    def groupId(self):
        return "vectorgeoprocessing"

    def commandName(self):
        return "ogr2ogr"

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        source = self.parameterAsSource(parameters, self.INPUT, context)
        if source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        fields = source.fields()
        input_details = self.getOgrCompatibleSource(
            self.INPUT, parameters, context, feedback, executing
        )

        if not input_details.layer_name:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        geometry = self.parameterAsString(parameters, self.GEOMETRY, context)
        if input_details.geometry_column_name:
            geometry = input_details.geometry_column_name

        fieldName = self.parameterAsString(parameters, self.FIELD, context)
        options = self.parameterAsString(parameters, self.OPTIONS, context)
        outFile = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)
        self.setOutputValue(self.OUTPUT, outFile)

        output_details = GdalUtils.gdal_connection_details_from_uri(outFile, context)

        other_fields_exist = any(True for f in fields if f.name() != geometry)

        other_fields = ",*" if other_fields_exist else ""

        arguments = [
            output_details.connection_string,
            input_details.connection_string,
            "-nlt PROMOTE_TO_MULTI",
            "-dialect",
            "sqlite",
            "-sql",
        ]

        tokens = []
        if self.parameterAsBoolean(parameters, self.COUNT_FEATURES, context):
            tokens.append(f"COUNT({geometry}) AS count")

        if self.parameterAsBoolean(parameters, self.COMPUTE_AREA, context):
            tokens.append(
                f"SUM(ST_Area({geometry})) AS area, ST_Perimeter(ST_Union({geometry})) AS perimeter"
            )

        statsField = self.parameterAsString(
            parameters, self.STATISTICS_ATTRIBUTE, context
        )
        if statsField and self.parameterAsBoolean(
            parameters, self.COMPUTE_STATISTICS, context
        ):
            tokens.append(
                f'SUM("{statsField}") AS sum, MIN("{statsField}") AS min, MAX("{statsField}") AS max, AVG("{statsField}") AS avg'
            )

        params = ",".join(tokens)
        if params:
            params = ", " + params

        group_by = ""
        if fieldName:
            group_by = f' GROUP BY "{fieldName}"'

        if self.parameterAsBoolean(parameters, self.KEEP_ATTRIBUTES, context):
            sql = f'SELECT ST_Union({geometry}) AS {geometry}{other_fields}{params} FROM "{input_details.layer_name}"{group_by}'
        else:
            sql = 'SELECT ST_Union({}) AS {}{}{} FROM "{}"{}'.format(
                geometry,
                geometry,
                f', "{fieldName}"' if fieldName else "",
                params,
                input_details.layer_name,
                group_by,
            )

        arguments.append(sql)

        if self.parameterAsBoolean(parameters, self.EXPLODE_COLLECTIONS, context):
            arguments.append("-explodecollections")

        if input_details.open_options:
            arguments.extend(input_details.open_options_as_arguments())

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        if options:
            arguments.append(options)

        if output_details.format:
            arguments.append(f"-f {output_details.format}")

        return [self.commandName(), GdalUtils.escapeAndJoin(arguments)]



"""
/***************************************************************************
Module to generate prepared APIs for calltips and auto-completion.
                             -------------------
begin                : 2012-09-10
copyright            : (C) 2012 Larry Shaffer
email                : larrys (at) dakotacarto (dot) com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
Portions of this file contain code from Eric4 APIsManager module.
"""

import os
from pathlib import Path

from qgis.PyQt import uic
from qgis.PyQt.Qsci import QsciAPIs, QsciLexerPython
from qgis.PyQt.QtCore import QCoreApplication
from qgis.PyQt.QtWidgets import QDialog, QDialogButtonBox

Ui_APIsDialogPythonConsole, _ = uic.loadUiType(
    Path(__file__).parent / "console_compile_apis.ui"
)


class PrepareAPIDialog(QDialog):
    def __init__(self, api_lexer, api_files, pap_file, parent=None):
        QDialog.__init__(self, parent)
        self.ui = Ui_APIsDialogPythonConsole()
        self.ui.setupUi(self)
        self.setWindowTitle(QCoreApplication.translate("PythonConsole", "Compile APIs"))
        self.ui.plainTextEdit.setVisible(False)
        self.ui.textEdit_Qsci.setVisible(False)
        self.adjustSize()
        self._api = None
        self.ui.buttonBox.rejected.connect(self._stopPreparation)
        self._api_files = api_files
        self._api_lexer = api_lexer
        self._pap_file = pap_file

    def _clearLexer(self):
        # self.ui.textEdit_Qsci.setLexer(0)
        self.qlexer = None

    def _stopPreparation(self):
        if self._api is not None:
            self._api.cancelPreparation()
        self._api = None
        self._clearLexer()
        self.close()

    def _preparationFinished(self):
        self._clearLexer()
        if os.path.exists(self._pap_file):
            os.remove(self._pap_file)
        self.ui.label.setText(
            QCoreApplication.translate("PythonConsole", "Saving prepared file…")
        )
        prepd = self._api.savePrepared(self._pap_file)
        rslt = self.tr("Error")
        if prepd:
            rslt = QCoreApplication.translate("PythonConsole", "Saved")
        self.ui.label.setText(f"{self.ui.label.text()} {rslt}")
        self._api = None
        self.ui.progressBar.setVisible(False)
        self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Cancel).setText(
            QCoreApplication.translate("PythonConsole", "Done")
        )
        self.adjustSize()

    def prepareAPI(self):
        # self.ui.textEdit_Qsci.setLexer(0)
        exec(f"self.qlexer = {self._api_lexer}(self.ui.textEdit_Qsci)")
        # self.ui.textEdit_Qsci.setLexer(self.qlexer)
        self._api = QsciAPIs(self.qlexer)
        self._api.apiPreparationFinished.connect(self._preparationFinished)
        for api_file in self._api_files:
            self._api.load(api_file)
        try:
            self._api.prepare()
        except Exception as err:
            self._api = None
            self._clearLexer()
            self.ui.label.setText(
                QCoreApplication.translate("PythonConsole", "Error preparing file…")
            )
            self.ui.progressBar.setVisible(False)
            self.ui.plainTextEdit.setVisible(True)
            self.ui.plainTextEdit.insertPlainText(err)
            self.ui.buttonBox.button(QDialogButtonBox.StandardButton.Cancel).setText(
                self.tr("Done")
            )
            self.adjustSize()



"""
***************************************************************************
    Relief.py
    ---------------------
    Date                 : December 2016
    Copyright            : (C) 2016 by Alexander Bruy
    Email                : alexander dot bruy at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Alexander Bruy"
__date__ = "December 2016"
__copyright__ = "(C) 2016, Alexander Bruy"

import os

from qgis.analysis import QgsRelief
from qgis.core import (
    QgsProcessingException,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterFileDestination,
    QgsProcessingParameterNumber,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsRasterFileWriter,
)
from qgis.PyQt.QtGui import QColor, QIcon

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class ParameterReliefColors(QgsProcessingParameterDefinition):
    def __init__(self, name="", description="", parent=None, optional=True):
        super().__init__(name, description, None, optional)
        self.parent = parent
        self.setMetadata(
            {
                "widget_wrapper": "processing.algs.qgis.ui.ReliefColorsWidget.ReliefColorsWidgetWrapper"
            }
        )

    def type(self):
        return "relief_colors"

    def clone(self):
        return ParameterReliefColors(
            self.name(),
            self.description(),
            self.parent,
            self.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional,
        )

    @staticmethod
    def valueToColors(value):
        if value is None:
            return None

        if value == "":
            return None

        if isinstance(value, str):
            return value.split(";")
        else:
            return ParameterReliefColors.colorsToString(value)

    @staticmethod
    def colorsToString(colors):
        return ";".join(
            f"{c[0]:f}, {c[1]:f}, {c[2]:d}, {c[3]:d}, {c[4]:d}" for c in colors
        )


class Relief(QgisAlgorithm):
    INPUT = "INPUT"
    Z_FACTOR = "Z_FACTOR"
    AUTO_COLORS = "AUTO_COLORS"
    COLORS = "COLORS"
    OUTPUT = "OUTPUT"
    FREQUENCY_DISTRIBUTION = "FREQUENCY_DISTRIBUTION"

    def icon(self):
        return QIcon(os.path.join(pluginPath, "images", "dem.png"))

    def group(self):
        return self.tr("Raster terrain analysis")

    def groupId(self):
        return "rasterterrainanalysis"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterRasterLayer(self.INPUT, self.tr("Elevation layer"))
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.Z_FACTOR,
                self.tr("Z factor"),
                type=QgsProcessingParameterNumber.Type.Double,
                minValue=0.00,
                defaultValue=1.0,
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.AUTO_COLORS,
                self.tr("Generate relief classes automatically"),
                defaultValue=False,
            )
        )
        self.addParameter(
            ParameterReliefColors(
                self.COLORS, self.tr("Relief colors"), self.INPUT, True
            )
        )
        self.addParameter(
            QgsProcessingParameterRasterDestination(self.OUTPUT, self.tr("Relief"))
        )
        self.addParameter(
            QgsProcessingParameterFileDestination(
                self.FREQUENCY_DISTRIBUTION,
                self.tr("Frequency distribution"),
                "CSV files (*.csv)",
                optional=True,
                createByDefault=False,
            )
        )

    def name(self):
        return "relief"

    def displayName(self):
        return self.tr("Relief")

    def shortDescription(self):
        return self.tr("Creates a shaded relief layer from digital elevation data.")

    def shortHelpString(self):
        return self.tr(
            "This algorithm creates a shaded relief layer from digital elevation data."
        )

    def processAlgorithm(self, parameters, context, feedback):
        inputFile = self.parameterAsRasterLayer(
            parameters, self.INPUT, context
        ).source()
        zFactor = self.parameterAsDouble(parameters, self.Z_FACTOR, context)
        automaticColors = self.parameterAsBoolean(parameters, self.AUTO_COLORS, context)
        outputFile = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)
        frequencyDistribution = self.parameterAsFileOutput(
            parameters, self.FREQUENCY_DISTRIBUTION, context
        )

        outputFormat = QgsRasterFileWriter.driverForExtension(
            os.path.splitext(outputFile)[1]
        )

        relief = QgsRelief(inputFile, outputFile, outputFormat)

        if automaticColors:
            reliefColors = relief.calculateOptimizedReliefClasses()
        else:
            colors = ParameterReliefColors.valueToColors(parameters[self.COLORS])
            if colors is None or len(colors) == 0:
                raise QgsProcessingException(
                    self.tr(
                        'Specify relief colors or activate "Generate relief classes automatically" option.'
                    )
                )

            reliefColors = []
            for c in colors:
                v = c.split(",")
                color = QgsRelief.ReliefColor(
                    QColor(int(v[2]), int(v[3]), int(v[4])), float(v[0]), float(v[1])
                )
                reliefColors.append(color)

        relief.setReliefColors(reliefColors)
        relief.setZFactor(zFactor)
        if frequencyDistribution:
            relief.exportFrequencyDistributionToCsv(frequencyDistribution)
        res = relief.processRaster(feedback)
        if res == 1:
            raise QgsProcessingException(self.tr("Can not open input file."))
        elif res == 2:
            raise QgsProcessingException(
                self.tr("Can not get GDAL driver for output file.")
            )
        elif res == 3:
            raise QgsProcessingException(self.tr("Can not create output file."))
        elif res == 4:
            raise QgsProcessingException(self.tr("Can not get input band."))
        elif res == 5:
            raise QgsProcessingException(self.tr("Can not create output bands."))
        elif res == 6:
            raise QgsProcessingException(
                self.tr("Output raster size is too small (at least 3 rows needed).")
            )
        elif res == 7:
            feedback.pushInfo(self.tr("Canceled."))

        return {
            self.OUTPUT: outputFile,
            self.FREQUENCY_DISTRIBUTION: frequencyDistribution,
        }



###############################################################################
#
# CSW Client
# ---------------------------------------------------------
# QGIS Catalog Service client.
#
# Copyright (C) 2010 NextGIS (http://nextgis.org),
#                    Alexander Bruy (alexander.bruy@gmail.com),
#                    Maxim Dubinin (sim@gis-lab.info)
#
# Copyright (C) 2014 Tom Kralidis (tomkralidis@gmail.com)
#
# This source is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This code is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
###############################################################################

import xml.etree.ElementTree as etree

from qgis.core import QgsSettings
from qgis.PyQt.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QListWidgetItem,
    QMessageBox,
)  # noqa

from MetaSearch.util import get_connections_from_file, get_ui_class, prettify_xml

BASE_CLASS = get_ui_class("manageconnectionsdialog.ui")


class ManageConnectionsDialog(QDialog, BASE_CLASS):
    """manage connections"""

    def __init__(self, mode):
        """init dialog"""

        QDialog.__init__(self)
        self.setupUi(self)
        self.settings = QgsSettings()
        self.filename = None
        self.mode = mode  # 0 - save, 1 - load
        self.btnBrowse.clicked.connect(self.select_file)
        self.manage_gui()

    def manage_gui(self):
        """manage interface"""

        if self.mode == 1:
            self.label.setText(self.tr("Load from file"))
            self.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setText(
                self.tr("Load")
            )
        else:
            self.label.setText(self.tr("Save to file"))
            self.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setText(
                self.tr("Save")
            )
            self.populate()

        self.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setEnabled(False)

    def select_file(self):
        """select file ops"""

        label = self.tr("eXtensible Markup Language (*.xml *.XML)")

        if self.mode == 0:
            slabel = self.tr("Save Connections")
            self.filename, filter = QFileDialog.getSaveFileName(
                self, slabel, ".", label
            )
        else:
            slabel = self.tr("Load Connections")
            self.filename, selected_filter = QFileDialog.getOpenFileName(
                self, slabel, ".", label
            )

        if not self.filename:
            return

        # ensure the user never omitted the extension from the file name
        if not self.filename.lower().endswith(".xml"):
            self.filename = "%s.xml" % self.filename

        self.leFileName.setText(self.filename)

        if self.mode == 1:
            self.populate()

        self.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setEnabled(True)

    def populate(self):
        """populate connections list from settings"""

        if self.mode == 0:
            self.settings.beginGroup("/MetaSearch/")
            keys = self.settings.childGroups()
            for key in keys:
                item = QListWidgetItem(self.listConnections)
                item.setText(key)
            self.settings.endGroup()

        else:  # populate connections list from file
            doc = get_connections_from_file(self, self.filename)
            if doc is None:
                self.filename = None
                self.leFileName.clear()
                self.listConnections.clear()
                return

            for catalog in doc.findall("csw"):
                item = QListWidgetItem(self.listConnections)
                item.setText(catalog.attrib.get("name"))

    def save(self, connections):
        """save connections ops"""

        doc = etree.Element("qgsCSWConnections")
        doc.attrib["version"] = "1.0"

        for conn in connections:
            url = self.settings.value("/MetaSearch/%s/url" % conn)
            type_ = self.settings.value("/MetaSearch/%s/catalog-type" % conn)
            if url is not None:
                connection = etree.SubElement(doc, "csw")
                connection.attrib["name"] = conn
                connection.attrib["type"] = type_ or "OGC CSW 2.0.2"
                connection.attrib["url"] = url

        # write to disk
        with open(self.filename, "w") as fileobj:
            fileobj.write(prettify_xml(etree.tostring(doc)))
        QMessageBox.information(
            self,
            self.tr("Save Connections"),
            self.tr("Saved to {0}.").format(self.filename),
        )
        self.reject()

    def load(self, items):
        """load connections"""

        self.settings.beginGroup("/MetaSearch/")
        keys = self.settings.childGroups()
        self.settings.endGroup()

        exml = etree.parse(self.filename).getroot()

        for catalog in exml.findall("csw"):
            conn_name = catalog.attrib.get("name")

            # process only selected connections
            if conn_name not in items:
                continue

            # check for duplicates
            if conn_name in keys:
                label = self.tr("File {0} exists. Overwrite?").format(conn_name)
                res = QMessageBox.warning(
                    self,
                    self.tr("Loading Connections"),
                    label,
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                )
                if res != QMessageBox.StandardButton.Yes:
                    continue

            # no dups detected or overwrite is allowed
            url = "/MetaSearch/%s/url" % conn_name
            type_ = "/MetaSearch/%s/catalog-type" % conn_name
            self.settings.setValue(url, catalog.attrib.get("url"))
            self.settings.setValue(
                type_, catalog.attrib.get("catalog-type", "OGC CSW 2.0.2")
            )

    def accept(self):
        """accept connections"""

        selection = self.listConnections.selectedItems()
        if len(selection) == 0:
            return

        items = []
        for sel in selection:
            items.append(sel.text())

        if self.mode == 0:  # save
            self.save(items)
        else:  # load
            self.load(items)

        self.filename = None
        self.leFileName.clear()
        self.listConnections.clear()
        self.buttonBox.button(QDialogButtonBox.StandardButton.Ok).setEnabled(False)

    def reject(self):
        """back out of manage connections dialogue"""

        QDialog.reject(self)



"""
***************************************************************************
    RenderingStyleFilePanel.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os
import warnings

from qgis.PyQt import uic
from qgis.PyQt.QtWidgets import QFileDialog

from processing.tools.system import isWindows

pluginPath = os.path.split(os.path.dirname(__file__))[0]

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    WIDGET, BASE = uic.loadUiType(
        os.path.join(pluginPath, "ui", "widgetBaseSelector.ui")
    )


class RenderingStyleFilePanel(BASE, WIDGET):
    def __init__(self):
        super().__init__(None)
        self.setupUi(self)

        self.btnSelect.clicked.connect(self.showSelectionDialog)

    def showSelectionDialog(self):
        filename, selected_filter = QFileDialog.getOpenFileName(
            self,
            self.tr("Select Style File"),
            "",
            self.tr("QGIS Layer Style File (*.qml *.QML)"),
        )
        if filename:
            self.leText.setText(filename)

    def setText(self, text):
        self.leText.setText(text)

    def getValue(self):
        s = self.leText.text()
        if isWindows():
            s = s.replace("\\", "/")
        return s



"""
***************************************************************************
    RandomSelectionWithinSubsets.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import random
from collections import defaultdict

from qgis.core import (
    QgsFeatureRequest,
    QgsFeatureSink,
    QgsProcessingException,
    QgsProcessingFeatureSource,
    QgsProcessingParameterEnum,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingParameterNumber,
)

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm


class RandomExtractWithinSubsets(QgisAlgorithm):
    INPUT = "INPUT"
    METHOD = "METHOD"
    NUMBER = "NUMBER"
    FIELD = "FIELD"
    OUTPUT = "OUTPUT"

    def group(self):
        return self.tr("Vector selection")

    def groupId(self):
        return "vectorselection"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.methods = [
            self.tr("Number of selected features"),
            self.tr("Percentage of selected features"),
        ]

        self.addParameter(
            QgsProcessingParameterFeatureSource(self.INPUT, self.tr("Input layer"))
        )

        self.addParameter(
            QgsProcessingParameterField(
                self.FIELD, self.tr("ID field"), None, self.INPUT
            )
        )

        self.addParameter(
            QgsProcessingParameterEnum(
                self.METHOD, self.tr("Method"), self.methods, False, 0
            )
        )

        self.addParameter(
            QgsProcessingParameterNumber(
                self.NUMBER,
                self.tr("Number/percentage of selected features"),
                QgsProcessingParameterNumber.Type.Integer,
                10,
                False,
                0.0,
            )
        )

        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT, self.tr("Extracted (random stratified)")
            )
        )

    def name(self):
        return "randomextractwithinsubsets"

    def displayName(self):
        return self.tr("Random extract within subsets")

    def shortDescription(self):
        return self.tr(
            "Generates a new vector layer that contains only a subset of the features in the input layer."
        )

    def shortHelpString(self):
        return self.tr(
            "This algorithm takes a vector layer and generates a new one "
            "that contains only a subset of the features in the input layer.\n"
            "The subset is defined randomly, using a percentage or count value "
            "to define the total number of features in the subset.\n"
            "The percentage/count value is not applied to the whole layer, but instead to each category. "
            "Categories are defined according to a given attribute, which is also specified as "
            "an input parameter for the algorithm."
        )

    def processAlgorithm(self, parameters, context, feedback):
        source = self.parameterAsSource(parameters, self.INPUT, context)
        if source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        method = self.parameterAsEnum(parameters, self.METHOD, context)

        field = self.parameterAsString(parameters, self.FIELD, context)

        index = source.fields().lookupField(field)

        features = source.getFeatures(
            QgsFeatureRequest(),
            QgsProcessingFeatureSource.Flag.FlagSkipGeometryValidityChecks,
        )
        featureCount = source.featureCount()
        unique = source.uniqueValues(index)
        value = self.parameterAsInt(parameters, self.NUMBER, context)
        if method == 0:
            if value > featureCount:
                raise QgsProcessingException(
                    self.tr(
                        "Selected number is greater that feature count. "
                        "Choose lesser value and try again."
                    )
                )
        else:
            if value > 100:
                raise QgsProcessingException(
                    self.tr(
                        "Percentage can't be greater than 100. Set "
                        "correct value and try again."
                    )
                )
            value = value / 100.0

        (sink, dest_id) = self.parameterAsSink(
            parameters,
            self.OUTPUT,
            context,
            source.fields(),
            source.wkbType(),
            source.sourceCrs(),
        )
        if sink is None:
            raise QgsProcessingException(self.invalidSinkError(parameters, self.OUTPUT))

        selran = []
        total = 100.0 / (featureCount * len(unique)) if featureCount else 1

        classes = defaultdict(list)
        for i, feature in enumerate(features):
            if feedback.isCanceled():
                break
            attrs = feature.attributes()
            classes[attrs[index]].append(feature)
            feedback.setProgress(int(i * total))

        for k, subset in classes.items():
            selValue = value if method != 1 else int(round(value * len(subset), 0))
            if selValue > len(subset):
                selValue = len(subset)
                feedback.reportError(
                    self.tr(
                        'Subset "{}" is smaller than requested number of features.'
                    ).format(k)
                )
            selran.extend(random.sample(subset, selValue))

        total = 100.0 / featureCount if featureCount else 1
        for i, feat in enumerate(selran):
            if feedback.isCanceled():
                break
            sink.addFeature(feat, QgsFeatureSink.Flag.FastInsert)
            feedback.setProgress(int(i * total))
        sink.finalize()
        return {self.OUTPUT: dest_id}



"""
***************************************************************************
    translate.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os

from qgis.core import (
    QgsProcessingException,
    QgsProcessingOutputRasterLayer,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterEnum,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterString,
)
from qgis.PyQt.QtGui import QIcon

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class gdaladdo(GdalAlgorithm):
    INPUT = "INPUT"
    LEVELS = "LEVELS"
    CLEAN = "CLEAN"
    RESAMPLING = "RESAMPLING"
    FORMAT = "FORMAT"
    EXTRA = "EXTRA"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.methods = (
            (self.tr("Nearest Neighbour (default)"), "nearest"),
            (self.tr("Average"), "average"),
            (self.tr("Gaussian"), "gauss"),
            (self.tr("Cubic (4x4 Kernel)"), "cubic"),
            (self.tr("Cubic B-Spline (4x4 Kernel)"), "cubicspline"),
            (self.tr("Lanczos (6x6 Kernel)"), "lanczos"),
            (self.tr("Average MP"), "average_mp"),
            (self.tr("Average in Mag/Phase Space"), "average_magphase"),
            (self.tr("Mode"), "mode"),
        )

        self.formats = (
            self.tr("Internal (if possible)"),
            self.tr("External (GTiff .ovr)"),
            self.tr("External (ERDAS Imagine .aux)"),
        )

        self.addParameter(
            QgsProcessingParameterRasterLayer(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.CLEAN, self.tr("Remove all existing overviews"), defaultValue=False
            )
        )

        if GdalUtils.version() < 230000:
            self.addParameter(
                QgsProcessingParameterString(
                    self.LEVELS, self.tr("Overview levels"), defaultValue="2 4 8 16"
                )
            )

        params = []
        if GdalUtils.version() >= 230000:
            params.append(
                QgsProcessingParameterString(
                    self.LEVELS,
                    self.tr("Overview levels (e.g. 2 4 8 16)"),
                    defaultValue=None,
                    optional=True,
                )
            )
        params.append(
            QgsProcessingParameterEnum(
                self.RESAMPLING,
                self.tr("Resampling method"),
                options=[i[0] for i in self.methods],
                allowMultiple=False,
                defaultValue=None,
                optional=True,
            )
        )
        params.append(
            QgsProcessingParameterEnum(
                self.FORMAT,
                self.tr("Overviews format"),
                options=self.formats,
                allowMultiple=False,
                defaultValue=0,
                optional=True,
            )
        )
        params.append(
            QgsProcessingParameterString(
                self.EXTRA,
                self.tr("Additional command-line parameters"),
                defaultValue=None,
                optional=True,
            )
        )
        for p in params:
            p.setFlags(p.flags() | QgsProcessingParameterDefinition.Flag.FlagAdvanced)
            self.addParameter(p)

        self.addOutput(
            QgsProcessingOutputRasterLayer(self.OUTPUT, self.tr("Pyramidized"))
        )

    def name(self):
        return "overviews"

    def displayName(self):
        return self.tr("Build overviews (pyramids)")

    def group(self):
        return self.tr("Raster miscellaneous")

    def groupId(self):
        return "rastermiscellaneous"

    def icon(self):
        return QIcon(
            os.path.join(pluginPath, "images", "gdaltools", "raster-overview.png")
        )

    def commandName(self):
        return "gdaladdo"

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        inLayer = self.parameterAsRasterLayer(parameters, self.INPUT, context)
        if inLayer is None:
            raise QgsProcessingException(
                self.invalidRasterError(parameters, self.INPUT)
            )
        input_details = GdalUtils.gdal_connection_details_from_layer(inLayer)

        arguments = [input_details.connection_string]
        if self.RESAMPLING in parameters and parameters[self.RESAMPLING] is not None:
            arguments.append("-r")
            arguments.append(
                self.methods[
                    self.parameterAsEnum(parameters, self.RESAMPLING, context)
                ][1]
            )

        ovrFormat = self.parameterAsEnum(parameters, self.FORMAT, context)
        if ovrFormat == 1:
            arguments.append("-ro")
        elif ovrFormat == 2:
            arguments.extend("--config USE_RRD YES".split(" "))

        if self.parameterAsBoolean(parameters, self.CLEAN, context):
            arguments.append("-clean")

        if self.EXTRA in parameters and parameters[self.EXTRA] not in (None, ""):
            extra = self.parameterAsString(parameters, self.EXTRA, context)
            arguments.append(extra)

        arguments.extend(
            self.parameterAsString(parameters, self.LEVELS, context).split(" ")
        )

        if input_details.credential_options:
            arguments.extend(input_details.credential_options_as_arguments())

        self.setOutputValue(self.OUTPUT, inLayer.source())

        return [self.commandName(), GdalUtils.escapeAndJoin(arguments)]



"""
***************************************************************************
    ProjectProvider.py
    ------------------------
    Date                 : July 2018
    Copyright            : (C) 2018 by Nyall Dawson
    Email                : nyall dot dawson at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Nyall Dawson"
__date__ = "July 2018"
__copyright__ = "(C) 2018, Nyall Dawson"

from qgis.core import (
    Qgis,
    QgsApplication,
    QgsMessageLog,
    QgsProcessingModelAlgorithm,
    QgsProcessingProvider,
    QgsProject,
    QgsRuntimeProfiler,
    QgsXmlUtils,
)

PROJECT_PROVIDER_ID = "project"


class ProjectProvider(QgsProcessingProvider):
    def __init__(self, project=None):
        super().__init__()
        if project is None:
            self.project = QgsProject.instance()
        else:
            self.project = project

        self.model_definitions = {}  # dict of models in project
        self.is_loading = False

        # must reload models if providers list is changed - previously unavailable algorithms
        # which models depend on may now be available
        QgsApplication.processingRegistry().providerAdded.connect(
            self.on_provider_added
        )

        self.project.readProject.connect(self.read_project)
        self.project.writeProject.connect(self.write_project)
        self.project.cleared.connect(self.clear)

    def on_provider_added(self, provider_id):
        if provider_id == self.id():
            return

        self.refreshAlgorithms()

    def load(self):
        with QgsRuntimeProfiler.profile("Project Provider"):
            self.refreshAlgorithms()

        return True

    def clear(self):
        """
        Remove all algorithms from the provider
        """
        self.model_definitions = {}
        self.refreshAlgorithms()

    def add_model(self, model):
        """
        Adds a model to the provider
        :type model: QgsProcessingModelAlgorithm
        :param model: model to add
        """
        definition = model.toVariant()
        self.model_definitions[model.name()] = definition
        self.refreshAlgorithms()

    def remove_model(self, model):
        """
        Removes a model from the project
        :type model: QgsProcessingModelAlgorithm
        :param model: model to remove
        """
        if model is None:
            return

        if model.name() in self.model_definitions:
            del self.model_definitions[model.name()]

        self.refreshAlgorithms()

    def read_project(self, doc):
        """
        Reads the project model definitions from the project DOM document
        :param doc: DOM document
        """
        self.model_definitions = {}
        project_models_nodes = doc.elementsByTagName("projectModels")
        if project_models_nodes:
            project_models_node = project_models_nodes.at(0)
            model_nodes = project_models_node.childNodes()
            for n in range(model_nodes.count()):
                model_element = model_nodes.at(n).toElement()
                definition = QgsXmlUtils.readVariant(model_element)
                algorithm = QgsProcessingModelAlgorithm()
                if algorithm.loadVariant(definition):
                    self.model_definitions[algorithm.name()] = definition

        self.refreshAlgorithms()

    def write_project(self, doc):
        """
        Writes out the project model definitions into the project DOM document
        :param doc: DOM document
        """
        qgis_nodes = doc.elementsByTagName("qgis")
        if not qgis_nodes:
            return

        qgis_node = qgis_nodes.at(0)
        project_models_node = doc.createElement("projectModels")

        for a in self.algorithms():
            definition = a.toVariant()
            element = QgsXmlUtils.writeVariant(definition, doc)
            project_models_node.appendChild(element)
        qgis_node.appendChild(project_models_node)

    def name(self):
        return self.tr("Project models", "ProjectProvider")

    def longName(self):
        return self.tr("Models embedded in the current project", "ProjectProvider")

    def id(self):
        return PROJECT_PROVIDER_ID

    def icon(self):
        return QgsApplication.getThemeIcon("/mIconQgsProjectFile.svg")

    def svgIconPath(self):
        return QgsApplication.iconPath("mIconQgsProjectFile.svg")

    def supportsNonFileBasedOutput(self):
        return True

    def loadAlgorithms(self):
        if self.is_loading:
            return
        self.is_loading = True

        for definition in self.model_definitions.values():
            algorithm = QgsProcessingModelAlgorithm()
            if algorithm.loadVariant(definition):
                self.addAlgorithm(algorithm)
            else:
                QgsMessageLog.logMessage(
                    self.tr("Could not load model from project", "ProjectProvider"),
                    self.tr("Processing"),
                    Qgis.MessageLevel.Critical,
                )

        self.is_loading = False



"""
***************************************************************************
    grass_algorithms_vector_test.py
    -----------------------------
    Date                 : April 2018
    Copyright            : (C) 2018 by Nyall Dawson
    Email                : nyall dot dawson at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Nyall Dawson"
__date__ = "March 2018"
__copyright__ = "(C) 2018, Nyall Dawson"

import os
import re
import shutil
import tempfile

import AlgorithmsTestBase
import nose2
from grassprovider.grass_provider import GrassProvider
from grassprovider.grass_utils import GrassUtils
from qgis.core import (
    QgsApplication,
    QgsFeature,
    QgsGeometry,
    QgsPointXY,
    QgsProcessingContext,
    QgsProcessingFeatureSourceDefinition,
    QgsProcessingFeedback,
    QgsProject,
    QgsVectorLayer,
)
from qgis.testing import QgisTestCase, start_app

testDataPath = os.path.join(os.path.dirname(__file__), "testdata")


class TestGrassAlgorithmsVectorTest(QgisTestCase, AlgorithmsTestBase.AlgorithmsTest):
    @classmethod
    def setUpClass(cls):
        start_app()
        cls.provider = GrassProvider()
        QgsApplication.processingRegistry().addProvider(cls.provider)
        cls.cleanup_paths = []

        cls.temp_dir = tempfile.mkdtemp()
        cls.cleanup_paths.append(cls.temp_dir)

        assert GrassUtils.installedVersion()

    @classmethod
    def tearDownClass(cls):
        QgsApplication.processingRegistry().removeProvider(cls.provider)
        for path in cls.cleanup_paths:
            shutil.rmtree(path)

    def definition_file(self):
        return "grass_algorithms_vector_tests.yaml"

    def testMemoryLayerInput(self):
        # create a memory layer and add to project and context
        layer = QgsVectorLayer(
            "Point?crs=epsg:3857&field=fldtxt:string&field=fldint:integer",
            "testmem",
            "memory",
        )
        self.assertTrue(layer.isValid())
        pr = layer.dataProvider()
        f = QgsFeature()
        f.setAttributes(["test", 123])
        f.setGeometry(QgsGeometry.fromPointXY(QgsPointXY(100, 200)))
        f2 = QgsFeature()
        f2.setAttributes(["test2", 457])
        f2.setGeometry(QgsGeometry.fromPointXY(QgsPointXY(110, 200)))
        self.assertTrue(pr.addFeatures([f, f2]))
        self.assertEqual(layer.featureCount(), 2)
        QgsProject.instance().addMapLayer(layer)
        context = QgsProcessingContext()
        context.setProject(QgsProject.instance())

        alg = QgsApplication.processingRegistry().createAlgorithmById("grass:v.buffer")
        self.assertIsNotNone(alg)

        temp_file = os.path.join(self.temp_dir, "grass_output.shp")
        parameters = {
            "input": "testmem",
            "cats": "",
            "where": "",
            "type": [0, 1, 4],
            "distance": 1,
            "minordistance": None,
            "angle": 0,
            "column": None,
            "scale": 1,
            "tolerance": 0.01,
            "-s": False,
            "-c": False,
            "-t": False,
            "output": temp_file,
            "GRASS_REGION_PARAMETER": None,
            "GRASS_SNAP_TOLERANCE_PARAMETER": -1,
            "GRASS_MIN_AREA_PARAMETER": 0.0001,
            "GRASS_OUTPUT_TYPE_PARAMETER": 0,
            "GRASS_VECTOR_DSCO": "",
            "GRASS_VECTOR_LCO": "",
        }
        feedback = QgsProcessingFeedback()

        results, ok = alg.run(parameters, context, feedback)
        self.assertTrue(ok)
        self.assertTrue(os.path.exists(temp_file))

        # make sure that layer has correct features
        res = QgsVectorLayer(temp_file, "res")
        self.assertTrue(res.isValid())
        self.assertEqual(res.featureCount(), 2)

        QgsProject.instance().removeMapLayer(layer)

    def testFeatureSourceInput(self):
        # create a memory layer and add to project and context
        layer = QgsVectorLayer(
            "Point?crs=epsg:3857&field=fldtxt:string&field=fldint:integer",
            "testmem",
            "memory",
        )
        self.assertTrue(layer.isValid())
        pr = layer.dataProvider()
        f = QgsFeature()
        f.setAttributes(["test", 123])
        f.setGeometry(QgsGeometry.fromPointXY(QgsPointXY(100, 200)))
        f2 = QgsFeature()
        f2.setAttributes(["test2", 457])
        f2.setGeometry(QgsGeometry.fromPointXY(QgsPointXY(110, 200)))
        self.assertTrue(pr.addFeatures([f, f2]))
        self.assertEqual(layer.featureCount(), 2)

        # select first feature
        layer.selectByIds([next(layer.getFeatures()).id()])
        self.assertEqual(len(layer.selectedFeatureIds()), 1)

        QgsProject.instance().addMapLayer(layer)
        context = QgsProcessingContext()
        context.setProject(QgsProject.instance())

        alg = QgsApplication.processingRegistry().createAlgorithmById("grass:v.buffer")
        self.assertIsNotNone(alg)
        temp_file = os.path.join(self.temp_dir, "grass_output_sel.shp")
        parameters = {
            "input": QgsProcessingFeatureSourceDefinition("testmem", True),
            "cats": "",
            "where": "",
            "type": [0, 1, 4],
            "distance": 1,
            "minordistance": None,
            "angle": 0,
            "column": None,
            "scale": 1,
            "tolerance": 0.01,
            "-s": False,
            "-c": False,
            "-t": False,
            "output": temp_file,
            "GRASS_REGION_PARAMETER": None,
            "GRASS_SNAP_TOLERANCE_PARAMETER": -1,
            "GRASS_MIN_AREA_PARAMETER": 0.0001,
            "GRASS_OUTPUT_TYPE_PARAMETER": 0,
            "GRASS_VECTOR_DSCO": "",
            "GRASS_VECTOR_LCO": "",
        }
        feedback = QgsProcessingFeedback()

        results, ok = alg.run(parameters, context, feedback)
        self.assertTrue(ok)
        self.assertTrue(os.path.exists(temp_file))

        # make sure that layer has correct features
        res = QgsVectorLayer(temp_file, "res")
        self.assertTrue(res.isValid())
        self.assertEqual(res.featureCount(), 1)

        QgsProject.instance().removeMapLayer(layer)

    def testOutputToGeopackage(self):
        # create a memory layer and add to project and context
        layer = QgsVectorLayer(
            "Point?crs=epsg:3857&field=fldtxt:string&field=fldint:integer",
            "testmem",
            "memory",
        )
        self.assertTrue(layer.isValid())
        pr = layer.dataProvider()
        f = QgsFeature()
        f.setAttributes(["test", 123])
        f.setGeometry(QgsGeometry.fromPointXY(QgsPointXY(100, 200)))
        f2 = QgsFeature()
        f2.setAttributes(["test2", 457])
        f2.setGeometry(QgsGeometry.fromPointXY(QgsPointXY(110, 200)))
        self.assertTrue(pr.addFeatures([f, f2]))
        self.assertEqual(layer.featureCount(), 2)
        QgsProject.instance().addMapLayer(layer)
        context = QgsProcessingContext()
        context.setProject(QgsProject.instance())

        alg = QgsApplication.processingRegistry().createAlgorithmById("grass:v.buffer")
        self.assertIsNotNone(alg)

        temp_file = os.path.join(self.temp_dir, "grass_output.gpkg")
        parameters = {
            "input": "testmem",
            "cats": "",
            "where": "",
            "type": [0, 1, 4],
            "distance": 1,
            "minordistance": None,
            "angle": 0,
            "column": None,
            "scale": 1,
            "tolerance": 0.01,
            "-s": False,
            "-c": False,
            "-t": False,
            "output": temp_file,
            "GRASS_REGION_PARAMETER": None,
            "GRASS_SNAP_TOLERANCE_PARAMETER": -1,
            "GRASS_MIN_AREA_PARAMETER": 0.0001,
            "GRASS_OUTPUT_TYPE_PARAMETER": 0,
            "GRASS_VECTOR_DSCO": "",
            "GRASS_VECTOR_LCO": "",
        }
        feedback = QgsProcessingFeedback()

        results, ok = alg.run(parameters, context, feedback)
        self.assertTrue(ok)
        self.assertTrue(os.path.exists(temp_file))

        # make sure that layer has correct features
        res = QgsVectorLayer(temp_file, "res")
        self.assertTrue(res.isValid())
        self.assertEqual(res.featureCount(), 2)

        QgsProject.instance().removeMapLayer(layer)

    def testVectorLayerInput(self):
        context = QgsProcessingContext()
        alg = QgsApplication.processingRegistry().createAlgorithmById("grass7:v.buffer")
        self.assertIsNotNone(alg)
        self.assertFalse(alg.commands)

        def get_command(alg):
            command = alg.commands[-1]
            command = re.sub(r'output=".*?"', 'output="###"', command)
            command = command.replace(testDataPath, "testdata")
            return command

        # GML source
        source = os.path.join(testDataPath, "points.gml")
        vl = QgsVectorLayer(source)
        self.assertTrue(vl.isValid())
        alg.loadVectorLayer("test_layer", vl, context, external=False)
        self.assertEqual(
            get_command(alg),
            'v.in.ogr min_area=None snap=None input="testdata/points.gml" output="###" --overwrite -o',
        )
        # try with external -- not support for GML, so should fall back to v.in.ogr
        alg.loadVectorLayer("test_layer", vl, context, external=True)
        self.assertEqual(
            get_command(alg),
            'v.in.ogr min_area=None snap=None input="testdata/points.gml" output="###" --overwrite -o',
        )

        # SHP source
        source = os.path.join(testDataPath, "lines_z.shp")
        vl = QgsVectorLayer(source)
        self.assertTrue(vl.isValid())
        alg.loadVectorLayer("test_layer", vl, context, external=False)
        self.assertEqual(
            get_command(alg),
            'v.in.ogr min_area=None snap=None input="testdata/lines_z.shp" output="###" --overwrite -o',
        )
        # try with external -- should work for shapefile
        alg.loadVectorLayer("test_layer", vl, context, external=True)
        self.assertEqual(
            get_command(alg),
            'v.external input="testdata/lines_z.shp" output="###" --overwrite -o',
        )

        # GPKG source
        source = os.path.join(testDataPath, "pol.gpkg")
        vl = QgsVectorLayer(source + "|layername=pol2")
        self.assertTrue(vl.isValid())
        alg.loadVectorLayer("test_layer", vl, context, external=False)
        self.assertEqual(
            get_command(alg),
            'v.in.ogr min_area=None snap=None input="testdata/pol.gpkg" layer="pol2" output="###" --overwrite -o',
        )
        # try with external -- should work for Geopackage (although grass itself tends to crash here!)
        alg.loadVectorLayer("test_layer", vl, context, external=True)
        self.assertEqual(
            get_command(alg),
            'v.external input="testdata/pol.gpkg" layer="pol2" output="###" --overwrite -o',
        )

        # different layer
        source = os.path.join(testDataPath, "pol.gpkg")
        vl = QgsVectorLayer(source + "|layername=pol3")
        self.assertTrue(vl.isValid())
        alg.loadVectorLayer("test_layer", vl, context, external=False)
        self.assertEqual(
            get_command(alg),
            'v.in.ogr min_area=None snap=None input="testdata/pol.gpkg" layer="pol3" output="###" --overwrite -o',
        )
        alg.loadVectorLayer("test_layer", vl, context, external=True)
        self.assertEqual(
            get_command(alg),
            'v.external input="testdata/pol.gpkg" layer="pol3" output="###" --overwrite -o',
        )

        # GPKG no layer: you get what you get and you don't get upset
        source = os.path.join(testDataPath, "pol.gpkg")
        vl = QgsVectorLayer(source)
        self.assertTrue(vl.isValid())
        alg.loadVectorLayer("test_layer", vl, context, external=False)
        self.assertEqual(
            get_command(alg),
            'v.in.ogr min_area=None snap=None input="testdata/pol.gpkg" output="###" --overwrite -o',
        )
        alg.loadVectorLayer("test_layer", vl, context, external=True)
        self.assertEqual(
            get_command(alg),
            'v.external input="testdata/pol.gpkg" output="###" --overwrite -o',
        )

        # layer with filter
        vl = QgsVectorLayer(source + "|layername=pol3")
        self.assertTrue(vl.isValid())
        vl.setSubsetString("\"field\"='value'")
        alg.loadVectorLayer("test_layer", vl, context, external=False)
        self.assertEqual(
            get_command(alg),
            'v.in.ogr min_area=None snap=None input="testdata/pol.gpkg" layer="pol3" output="###" --overwrite -o where="\\"field\\"=\'value\'"',
        )
        alg.loadVectorLayer("test_layer", vl, context, external=True)
        self.assertEqual(
            get_command(alg),
            'v.external input="testdata/pol.gpkg" layer="pol3" output="###" --overwrite -o where="\\"field\\"=\'value\'"',
        )


if __name__ == "__main__":
    nose2.main()



"""
***************************************************************************
    qgsfunction.py
    ---------------------
    Date                 : May 2018
    Copyright            : (C) 2018 by Denis Rouzaud
    Email                : denis@opengis.ch
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

import inspect
import string
import traceback

from qgis._core import (
    Qgis,
    QgsExpression,
    QgsExpressionFunction,
    QgsFeatureRequest,
    QgsMessageLog,
)
from qgis.PyQt.QtCore import QCoreApplication


class QgsPyExpressionFunction(QgsExpressionFunction):
    """Python expression function"""

    def __init__(
        self,
        function,
        name,
        group,
        helptext="",
        usesgeometry=False,
        referenced_columns=QgsFeatureRequest.ALL_ATTRIBUTES,
        handlesnull=False,
        params_as_list=False,
    ):
        # Call the parent constructor
        # -1 means that function can take any number of arguments
        QgsExpressionFunction.__init__(self, name, -1, group, helptext)
        self.function = function
        self.params_as_list = params_as_list
        self.usesgeometry = usesgeometry
        self.referenced_columns = referenced_columns
        self.handlesnull = handlesnull

    def func(self, values, context, parent, node):
        feature = None
        if context:
            feature = context.feature()
        try:
            # Inspect the inner function signature to get the list of parameters
            parameters = inspect.signature(self.function).parameters
            kwvalues = {}

            # Handle special parameters
            # those will not be inserted in the arguments list
            # if they are present in the function signature
            if "context" in parameters:
                kwvalues["context"] = context
            if "feature" in parameters:
                kwvalues["feature"] = feature
            if "parent" in parameters:
                kwvalues["parent"] = parent

            # In this context, values is a list of the parameters passed to the expression.
            # If self.params_as_list is True, values is passed as is to the inner function.
            if self.params_as_list:
                return self.function(values, **kwvalues)
            # Otherwise (default), the parameters are expanded
            return self.function(*values, **kwvalues)

        except Exception as ex:
            tb = traceback.format_exception(None, ex, ex.__traceback__)
            formatted_traceback = "".join(tb)
            formatted_exception = f"{ex}:<pre>{formatted_traceback}</pre>"
            parent.setEvalErrorString(formatted_exception)
            return None

    def usesGeometry(self, node):
        return self.usesgeometry

    def referencedColumns(self, node):
        return self.referenced_columns

    def handlesNull(self):
        return self.handlesnull


def register_function(
    function,
    args="auto",
    group: str = "custom",
    usesgeometry: bool = False,
    referenced_columns=[QgsFeatureRequest.ALL_ATTRIBUTES],
    handlesnull: bool = False,
    params_as_list: bool = False,
    **kwargs,
):
    r"""
    Register a Python function to be used as a expression function.

    The function signature may contain special parameters (in any order at the end of the signature):

    - feature: the QgsFeature related to the current evaluation
    - parent: the QgsExpression parent
    - context: the QgsExpressionContext related to the current evaluation

    If those parameters are present in the function signature, they will be automatically passed to the function,
    without the need to specify them in the expression.

    Functions should return a value compatible with QVariant.

    Evaluation errors can be raised using:

    .. code-block:: python

        parent.setEvalErrorString("Error message")

    :param function: the Python function to be used as an expression function
    :param args: Argument list

      .. deprecated:: 3.32
          Use ``params_as_list`` if you want to pass parameters as a list.

    :param group: the expression group in which the function should be added
    :param usesgeometry: Defines if this expression requires the geometry. By default ``False``.
    :param referenced_columns: An array of names of fields on which this expression works. By default ``[QgsFeatureRequest.ALL_ATTRIBUTES]``. Specifying a subset of fields or an empty list will result in a faster execution.
    :param handlesnull: Defines if this expression has custom handling for NULL values. If ``False``, the result will always be NULL as soon as any parameter is NULL. ``False`` by default.
    :param params_as_list: If True, the function will receive the expression parameters as a list. If ``False``, the function will receive the parameters as individual arguments. ``False`` by default.
    :param \**kwargs:
        See below
    :Keyword Arguments:

    * *register* (``bool``): Set to ``False`` to create the QgsPyExpressionFunction without registering it. Useful for testing puposes. By default ``True``.
    * *name* (``str``): If provided, replace the function name
    * *helpText* (``str``): If provided, used in the help tooltip instead of the function docstring

    :return:
    """

    # Format the help text
    helptemplate = string.Template("<h3>function $name</h3>\n$doc")
    name = kwargs.get("name", function.__name__)
    helptext = kwargs.get("helpText") or function.__doc__ or ""
    helptext = helptext.strip()

    register = kwargs.get("register", True)
    if register and QgsExpression.isFunctionName(name):
        if not QgsExpression.unregisterFunction(name):
            msgtitle = QCoreApplication.translate("UserExpressions", "User expressions")
            msg = QCoreApplication.translate(
                "UserExpressions",
                "The user expression {0} already exists and could not be unregistered.",
            ).format(name)
            QgsMessageLog.logMessage(msg + "\n", msgtitle, Qgis.MessageLevel.Warning)
            return None

    function.__name__ = name
    helptext = helptemplate.safe_substitute(name=name, doc=helptext)

    # Legacy: if args was not 'auto', parameters were passed as a list
    params_as_list = params_as_list or args != "auto"
    f = QgsPyExpressionFunction(
        function,
        name,
        group,
        helptext,
        usesgeometry,
        referenced_columns,
        handlesnull,
        params_as_list,
    )

    if register:
        QgsExpression.registerFunction(f)
    return f


def qgsfunction(args="auto", group: str = "custom", **kwargs):
    r"""
    Decorator function used to define a user expression function.

    The decorated function signature may contain special parameters (in any order at the end of the signature):

    - ``feature``: the QgsFeature related to the current evaluation
    - ``parent``: the QgsExpression parent
    - ``context``: the QgsExpressionContext related to the current evaluation

    If those parameters are present in the function signature, they will be automatically passed to the function,
    without the need to specify them in the expression.

    Functions should return a value compatible with QVariant.

    Evaluation errors can be raised using:

    .. code-block:: python

        parent.setEvalErrorString("Error message")

    :param args: Argument list

      .. deprecated:: 3.32
          Use ``params_as_list`` if you want to pass parameters as a list.

    :param str group: The expression group to which this expression should be added.
    :param \**kwargs:
        See below
    :Keyword Arguments:
        * **usesgeometry** (*bool*) --
          Defines if this expression requires the geometry. By default ``False``.
        * **referenced_columns** (*List[str]*) --
          An array of names of fields on which this expression works. By default ``[QgsFeatureRequest.ALL_ATTRIBUTES]``. Specifying a subset of fields or an empty list will result in a faster execution.
        * **handlesnull** (*bool*) --
          Defines if this expression has custom handling for NULL values. If ``False``, the result will always be NULL as soon as any parameter is NULL. ``False`` by default.
        * **params_as_list** (*bool*) --
          Defines if the parameters are passed to the function as a list. If set to ``False``, they will be expanded. By default ``False``. Since QGIS 3.32.
        * **register** (*bool*) --
          Set to ``False`` to create the QgsPyExpressionFunction without registering it. Useful for testing puposes. By default ``True``.
        * **name** (*str*) --
          If provided, replace the function name.
        * **helpText** (*str*) --
          If provided, used in the help tooltip instead of the function docstring.


    Example 1 (Basic function, with default parameters)
    ---------------------------------------------------

    .. code-block:: python

        @qgsfunction(group="python")
        def center(text, width, fillchar=" "):
            return text.center(width, fillchar)

    This registers a function called "center" in the "python" group.
    This expression requires two parameters: ``text`` and ``width``.
    An optional third parameter can be provided: ``fillchar``.

    .. code-block:: text

        >>> center('Hello', 10)
        '  Hello   '
        >>> center('Hello', 15, '*')
        '*****Hello*****'


    Example 2 (variable number of parameters)
    -----------------------------------------

    .. code-block:: python

        @qgsfunction(group="custom")
        def fibonnaci_numbers(*args):
            def fibonnaci(n):
                if n <= 1:
                    return n
                else:
                    return fibonnaci(n - 1) + fibonnaci(n - 2)
            return [fibonnaci(arg) for arg in args]

    This registers a function called "fibonnaci_numbers" in the "custom" group.
    This expression can be called with any number of parameters, and will return a list of the fibonnaci numbers
    corresponding to the parameters.

    .. code-block:: text

        >>> fibonnaci_numbers()
        []
        >>> fibonnaci_numbers(1)
        [1]
        >>> fibonnaci_numbers(3)
        [2]
        >>> fibonnaci_numbers(1, 2, 3, 4, 5, 6, 7)
        [ 1, 1, 2, 3, 5, 8, 13 ]


    Example 3 (feature and parent)
    ------------------------------

    .. code-block:: python

        @qgsfunction(group="custom")
        def display_field(fieldname, feature, parent):
            try:
                return f"<b>{fieldname}</b>: {feature[fieldname]}"
            except KeyError:
                parent.setEvalErrorString(f"Field {fieldname} not found")

    This registers a function called "display_field" in the "custom" group.
    This expression requires an unique parameter: ``fieldname``. ``feature`` is automatically passed to the function.
    ``parent`` is the QgsExpression parent, and can be used to raise an error.

    .. code-block:: text

        >>> display_field("altitude")
        <b>altitude</b>: 164
        >>> display_field("lat")
        <b>lat</b>: 45.4

    Example 4 (context)
    -------------------

    .. code-block:: python

        @qgsfunction(group="custom")
        def title_layer_name(context):
            return context.variable("layer_name").title()

    This registers a function called "title_layer_name" in the "custom" group. It takes no parameters,
    but extracts the layer name from the expression context and returns it as a title case string.

    """

    def wrapper(func):
        return register_function(func, args, group, **kwargs)

    return wrapper



"""
***************************************************************************
    dataobject.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import os
import re

from qgis.core import (
    QgsDataProvider,
    QgsExpressionContext,
    QgsExpressionContextScope,
    QgsExpressionContextUtils,
    QgsFeatureRequest,
    QgsProcessingContext,
    QgsProcessingUtils,
    QgsProject,
    QgsRasterLayer,
    QgsSettings,
    QgsVectorLayer,
    QgsWkbTypes,
)
from qgis.gui import QgsSublayersDialog
from qgis.PyQt.QtCore import QCoreApplication
from qgis.utils import iface

from processing.core.ProcessingConfig import ProcessingConfig

ALL_TYPES = [-1]

TYPE_VECTOR_ANY = -1
TYPE_VECTOR_POINT = 0
TYPE_VECTOR_LINE = 1
TYPE_VECTOR_POLYGON = 2
TYPE_RASTER = 3
TYPE_FILE = 4
TYPE_TABLE = 5


# changing this signature? make sure you update the signature in
# python/processing/__init__.py too!
# Docstring for this function is in python/processing/__init__.py
def createContext(feedback=None):
    context = QgsProcessingContext()
    context.setProject(QgsProject.instance())
    context.setFeedback(feedback)

    invalid_features_method = ProcessingConfig.getSetting(
        ProcessingConfig.FILTER_INVALID_GEOMETRIES
    )
    if invalid_features_method is None:
        invalid_features_method = (
            QgsFeatureRequest.InvalidGeometryCheck.GeometryAbortOnInvalid
        )
    else:
        invalid_features_method = QgsFeatureRequest.InvalidGeometryCheck(
            int(invalid_features_method)
        )
    context.setInvalidGeometryCheck(invalid_features_method)

    settings = QgsSettings()
    context.setDefaultEncoding(
        QgsProcessingUtils.resolveDefaultEncoding(
            settings.value("/Processing/encoding")
        )
    )

    context.setExpressionContext(createExpressionContext())

    if iface and iface.mapCanvas() and iface.mapCanvas().mapSettings().isTemporal():
        context.setCurrentTimeRange(iface.mapCanvas().mapSettings().temporalRange())

    return context


def createExpressionContext():
    context = QgsExpressionContext()
    context.appendScope(QgsExpressionContextUtils.globalScope())
    context.appendScope(QgsExpressionContextUtils.projectScope(QgsProject.instance()))

    if iface and iface.mapCanvas():
        context.appendScope(
            QgsExpressionContextUtils.mapSettingsScope(iface.mapCanvas().mapSettings())
        )

    processingScope = QgsExpressionContextScope()

    if iface and iface.mapCanvas():
        extent = iface.mapCanvas().fullExtent()
        processingScope.setVariable("fullextent_minx", extent.xMinimum())
        processingScope.setVariable("fullextent_miny", extent.yMinimum())
        processingScope.setVariable("fullextent_maxx", extent.xMaximum())
        processingScope.setVariable("fullextent_maxy", extent.yMaximum())

    context.appendScope(processingScope)
    return context


def load(fileName, name=None, crs=None, style=None, isRaster=False):
    """
    Loads a layer/table into the current project, given its file.

    .. deprecated:: 3.0
    Do not use, will be removed in QGIS 5.0
    """

    from warnings import warn

    warn(
        "processing.load is deprecated and will be removed in QGIS 5.0",
        DeprecationWarning,
    )

    if fileName is None:
        return

    if name is None:
        name = os.path.split(fileName)[1]

    if isRaster:
        options = QgsRasterLayer.LayerOptions()
        options.skipCrsValidation = True
        qgslayer = QgsRasterLayer(fileName, name, "gdal", options)
        if qgslayer.isValid():
            if crs is not None and qgslayer.crs() is None:
                qgslayer.setCrs(crs, False)
            if style is None:
                style = ProcessingConfig.getSetting(ProcessingConfig.RASTER_STYLE)
            qgslayer.loadNamedStyle(style)
            QgsProject.instance().addMapLayers([qgslayer])
        else:
            raise RuntimeError(
                QCoreApplication.translate(
                    "dataobject",
                    "Could not load layer: {0}\nCheck the processing framework log to look for errors.",
                ).format(fileName)
            )
    else:
        options = QgsVectorLayer.LayerOptions()
        options.skipCrsValidation = True
        qgslayer = QgsVectorLayer(fileName, name, "ogr", options)
        if qgslayer.isValid():
            if crs is not None and qgslayer.crs() is None:
                qgslayer.setCrs(crs, False)
            if style is None:
                if qgslayer.geometryType() == QgsWkbTypes.GeometryType.PointGeometry:
                    style = ProcessingConfig.getSetting(
                        ProcessingConfig.VECTOR_POINT_STYLE
                    )
                elif qgslayer.geometryType() == QgsWkbTypes.GeometryType.LineGeometry:
                    style = ProcessingConfig.getSetting(
                        ProcessingConfig.VECTOR_LINE_STYLE
                    )
                else:
                    style = ProcessingConfig.getSetting(
                        ProcessingConfig.VECTOR_POLYGON_STYLE
                    )
            qgslayer.loadNamedStyle(style)
            QgsProject.instance().addMapLayers([qgslayer])

    return qgslayer


def getRasterSublayer(path, param):
    layer = QgsRasterLayer(path)

    try:
        # If the layer is a raster layer and has multiple sublayers, let the user chose one.
        # Based on QgisApp::askUserForGDALSublayers
        if (
            layer
            and param.showSublayersDialog
            and layer.dataProvider().name() == "gdal"
            and len(layer.subLayers()) > 1
        ):
            layers = []
            subLayerNum = 0
            # simplify raster sublayer name
            for subLayer in layer.subLayers():
                # if netcdf/hdf use all text after filename
                if bool(re.match("netcdf", subLayer, re.I)) or bool(
                    re.match("hdf", subLayer, re.I)
                ):
                    subLayer = subLayer.split(path)[1]
                    subLayer = subLayer[1:]
                else:
                    # remove driver name and file name
                    subLayer.replace(
                        subLayer.split(QgsDataProvider.SUBLAYER_SEPARATOR)[0], ""
                    )
                    subLayer.replace(path, "")
                # remove any : or " left over
                if subLayer.startswith(":"):
                    subLayer = subLayer[1:]
                if subLayer.startswith('"'):
                    subLayer = subLayer[1:]
                if subLayer.endswith(":"):
                    subLayer = subLayer[:-1]
                if subLayer.endswith('"'):
                    subLayer = subLayer[:-1]

                ld = QgsSublayersDialog.LayerDefinition()
                ld.layerId = subLayerNum
                ld.layerName = subLayer
                layers.append(ld)
                subLayerNum = subLayerNum + 1

            # Use QgsSublayersDialog
            # Would be good if QgsSublayersDialog had an option to allow only one sublayer to be selected
            chooseSublayersDialog = QgsSublayersDialog(
                QgsSublayersDialog.ProviderType.Gdal, "gdal"
            )
            chooseSublayersDialog.populateLayerTable(layers)

            if chooseSublayersDialog.exec():
                return layer.subLayers()[chooseSublayersDialog.selectionIndexes()[0]]
            else:
                # If user pressed cancel then just return the input path
                return path
        else:
            # If the sublayers selection dialog is not to be shown then just return the input path
            return path
    except:
        # If the layer is not a raster layer, then just return the input path
        return path



"""
***************************************************************************
    v_rectify.py
    ------------
    Date                 : March 2016
    Copyright            : (C) 2016 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "March 2016"
__copyright__ = "(C) 2016, Médéric Ribreux"

import os

from processing.tools.system import getTempFilename
from qgis.PyQt.QtCore import QCoreApplication

from grassprovider.grass_utils import GrassUtils


def checkParameterValuesBeforeExecuting(alg, parameters, context):
    """Verify if we have the right parameters"""
    if alg.parameterAsString(
        parameters, "inline_points", context
    ) and alg.parameterAsString(parameters, "points", context):
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt",
            "You need to set either an input control point file or inline control points!",
        )

    return True, None


def processCommand(alg, parameters, context, feedback):
    # handle inline points
    inlinePoints = alg.parameterAsString(parameters, "inline_points", context)
    if inlinePoints:
        # Creates a temporary txt file
        pointsName = getTempFilename(context=context)

        # Inject rules into temporary txt file
        with open(pointsName, "w") as tempPoints:
            tempPoints.write(inlinePoints)
        alg.removeParameter("inline_points")
        parameters["points"] = pointsName

    alg.processCommand(parameters, context, feedback)



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : Oct 13, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

The content of this file is based on
- PG_Manager by Martin Dobias (GPLv2 license)
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.PyQt import uic
from qgis.PyQt.QtCore import QItemSelectionModel, QModelIndex, Qt, pyqtSignal
from qgis.PyQt.QtWidgets import (
    QApplication,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QItemDelegate,
    QMessageBox,
    QPushButton,
)
from qgis.utils import OverrideCursor

from .db_plugins.data_model import TableFieldsModel
from .db_plugins.plugin import ConnectionError, DbError
from .dlg_db_error import DlgDbError
from .gui_utils import GuiUtils

Ui_Dialog, _ = uic.loadUiType(GuiUtils.get_ui_file_path("DlgCreateTable.ui"))


class TableFieldsDelegate(QItemDelegate):
    """delegate with some special item editors"""

    columnNameChanged = pyqtSignal()

    def __init__(self, field_types, parent=None):
        QItemDelegate.__init__(self, parent)
        self.fieldTypes = field_types

    def createEditor(self, parent, option, index):
        # special combobox for field type
        if index.column() == 1:
            cbo = QComboBox(parent)
            cbo.setEditable(True)
            cbo.setFrame(False)
            for item in self.fieldTypes:
                cbo.addItem(item)
            return cbo
        return QItemDelegate.createEditor(self, parent, option, index)

    def setEditorData(self, editor, index):
        """load data from model to editor"""
        m = index.model()
        if index.column() == 1:
            txt = m.data(index, Qt.ItemDataRole.DisplayRole)
            editor.setEditText(txt)
        else:
            # use default
            QItemDelegate.setEditorData(self, editor, index)

    def setModelData(self, editor, model, index):
        """save data from editor back to model"""
        if index.column() == 1:
            model.setData(index, editor.currentText())
        else:
            # use default
            QItemDelegate.setModelData(self, editor, model, index)
            if index.column() == 0:
                self.columnNameChanged.emit()


class DlgCreateTable(QDialog, Ui_Dialog):
    GEOM_TYPES = [
        "POINT",
        "LINESTRING",
        "POLYGON",
        "MULTIPOINT",
        "MULTILINESTRING",
        "MULTIPOLYGON",
        "GEOMETRYCOLLECTION",
    ]

    def __init__(self, item, parent=None):
        QDialog.__init__(self, parent)
        self.item = item
        self.setupUi(self)

        self.db = self.item.database()
        self.schemas = self.db.schemas()
        self.hasSchemas = self.schemas is not None
        self.fieldTypes = self.db.connector.fieldTypes()

        m = TableFieldsModel(self, True)  # it's editable
        self.fields.setModel(m)
        self.fields.setColumnHidden(3, True)  # hide Default column

        d = TableFieldsDelegate(self.fieldTypes, self)
        self.fields.setItemDelegate(d)

        self.fields.setColumnWidth(0, 140)
        self.fields.setColumnWidth(1, 140)
        self.fields.setColumnWidth(2, 50)

        b = QPushButton(self.tr("&Create"))
        self.buttonBox.addButton(b, QDialogButtonBox.ButtonRole.ActionRole)

        self.btnAddField.clicked.connect(self.addField)
        self.btnDeleteField.clicked.connect(self.deleteField)
        self.btnFieldUp.clicked.connect(self.fieldUp)
        self.btnFieldDown.clicked.connect(self.fieldDown)
        b.clicked.connect(self.createTable)

        self.chkGeomColumn.clicked.connect(self.updateUi)

        self.fields.selectionModel().selectionChanged.connect(self.updateUiFields)
        d.columnNameChanged.connect(self.updatePkeyCombo)

        self.populateSchemas()
        self.updateUi()
        self.updateUiFields()

    def populateSchemas(self):
        self.cboSchema.clear()
        if not self.hasSchemas:
            self.hideSchemas()
            return

        index = -1
        for schema in self.schemas:
            self.cboSchema.addItem(schema.name)
            if hasattr(self.item, "schema") and schema.name == self.item.schema().name:
                index = self.cboSchema.count() - 1
        self.cboSchema.setCurrentIndex(index)

    def hideSchemas(self):
        self.cboSchema.setEnabled(False)

    def updateUi(self):
        useGeom = self.chkGeomColumn.isChecked()
        self.cboGeomType.setEnabled(useGeom)
        self.editGeomColumn.setEnabled(useGeom)
        self.spinGeomDim.setEnabled(useGeom)
        self.editGeomSrid.setEnabled(useGeom)
        self.chkSpatialIndex.setEnabled(useGeom)

    def updateUiFields(self):
        fld = self.selectedField()
        if fld is not None:
            up_enabled = fld != 0
            down_enabled = fld != self.fields.model().rowCount() - 1
            del_enabled = True
        else:
            up_enabled, down_enabled, del_enabled = False, False, False
        self.btnFieldUp.setEnabled(up_enabled)
        self.btnFieldDown.setEnabled(down_enabled)
        self.btnDeleteField.setEnabled(del_enabled)

    def updatePkeyCombo(self, selRow=None):
        """called when list of columns changes. if 'sel' is None, it keeps current index"""

        if selRow is None:
            selRow = self.cboPrimaryKey.currentIndex()

        self.cboPrimaryKey.clear()

        m = self.fields.model()
        for row in range(m.rowCount()):
            name = m.data(m.index(row, 0))
            self.cboPrimaryKey.addItem(name)

        self.cboPrimaryKey.setCurrentIndex(selRow)

    def addField(self):
        """Adds new field to the end of field table"""
        m = self.fields.model()
        newRow = m.rowCount()
        m.insertRows(newRow, 1)

        indexName = m.index(newRow, 0, QModelIndex())
        indexType = m.index(newRow, 1, QModelIndex())
        indexNull = m.index(newRow, 2, QModelIndex())

        m.setData(indexName, "new_field")
        colType = self.fieldTypes[0]
        if newRow == 0:
            # adding the first row, use auto-incrementing column type if any
            if "serial" in self.fieldTypes:  # PostgreSQL
                colType = "serial"
        m.setData(indexType, colType)
        m.setData(indexNull, None, Qt.ItemDataRole.DisplayRole)
        m.setData(indexNull, Qt.CheckState.Unchecked, Qt.ItemDataRole.CheckStateRole)

        # selects the new row
        sel = self.fields.selectionModel()
        sel.select(
            indexName,
            QItemSelectionModel.SelectionFlag.Rows
            | QItemSelectionModel.SelectionFlag.ClearAndSelect,
        )

        # starts editing
        self.fields.edit(indexName)

        self.updatePkeyCombo(0 if newRow == 0 else None)

    def selectedField(self):
        sel = self.fields.selectedIndexes()
        if len(sel) < 1:
            return None
        return sel[0].row()

    def deleteField(self):
        """Deletes selected field"""
        row = self.selectedField()
        if row is None:
            QMessageBox.information(
                self, self.tr("DB Manager"), self.tr("No field selected.")
            )
        else:
            self.fields.model().removeRows(row, 1)

        self.updatePkeyCombo()

    def fieldUp(self):
        """move selected field up"""
        row = self.selectedField()
        if row is None:
            QMessageBox.information(
                self, self.tr("DB Manager"), self.tr("No field selected.")
            )
            return
        if row == 0:
            QMessageBox.information(
                self, self.tr("DB Manager"), self.tr("Field is already at the top.")
            )
            return

        # take row and reinsert it
        rowdata = self.fields.model().takeRow(row)
        self.fields.model().insertRow(row - 1, rowdata)

        # set selection again
        index = self.fields.model().index(row - 1, 0, QModelIndex())
        self.fields.selectionModel().select(
            index,
            QItemSelectionModel.SelectionFlag.Rows
            | QItemSelectionModel.SelectionFlag.ClearAndSelect,
        )

        self.updatePkeyCombo()

    def fieldDown(self):
        """move selected field down"""
        row = self.selectedField()
        if row is None:
            QMessageBox.information(
                self, self.tr("DB Manager"), self.tr("No field selected.")
            )
            return
        if row == self.fields.model().rowCount() - 1:
            QMessageBox.information(
                self, self.tr("DB Manager"), self.tr("Field is already at the bottom.")
            )
            return

        # take row and reinsert it
        rowdata = self.fields.model().takeRow(row)
        self.fields.model().insertRow(row + 1, rowdata)

        # set selection again
        index = self.fields.model().index(row + 1, 0, QModelIndex())
        self.fields.selectionModel().select(
            index,
            QItemSelectionModel.SelectionFlag.Rows
            | QItemSelectionModel.SelectionFlag.ClearAndSelect,
        )

        self.updatePkeyCombo()

    def createTable(self):
        """Creates table with chosen fields, optionally add a geometry column"""
        if not self.hasSchemas:
            schema = None
        else:
            schema = str(self.cboSchema.currentText())
            if len(schema) == 0:
                QMessageBox.information(
                    self,
                    self.tr("DB Manager"),
                    self.tr("A valid schema must be selected first."),
                )
                return

        table = str(self.editName.text())
        if len(table) == 0:
            QMessageBox.information(
                self, self.tr("DB Manager"), self.tr("A valid table name is required.")
            )
            return

        m = self.fields.model()
        if m.rowCount() == 0:
            QMessageBox.information(
                self, self.tr("DB Manager"), self.tr("At least one field is required.")
            )
            return

        useGeomColumn = self.chkGeomColumn.isChecked()
        if useGeomColumn:
            geomColumn = str(self.editGeomColumn.text())
            if len(geomColumn) == 0:
                QMessageBox.information(
                    self,
                    self.tr("DB Manager"),
                    self.tr("A name is required for the geometry column."),
                )
                return

            geomType = self.GEOM_TYPES[self.cboGeomType.currentIndex()]
            geomDim = self.spinGeomDim.value()
            try:
                geomSrid = int(self.editGeomSrid.text())
            except ValueError:
                geomSrid = 0
            useSpatialIndex = self.chkSpatialIndex.isChecked()

        flds = m.getFields()
        pk_index = self.cboPrimaryKey.currentIndex()
        if pk_index >= 0:
            flds[pk_index].primaryKey = True

        # commit to DB
        with OverrideCursor(Qt.CursorShape.WaitCursor):
            try:
                if not useGeomColumn:
                    self.db.createTable(table, flds, schema)
                else:
                    geom = geomColumn, geomType, geomSrid, geomDim, useSpatialIndex
                    self.db.createVectorTable(table, flds, geom, schema)

            except (ConnectionError, DbError) as e:
                DlgDbError.showError(e, self)

        # clear UI
        self.editName.clear()
        self.fields.model().removeRows(0, self.fields.model().rowCount())
        self.cboPrimaryKey.clear()
        self.chkGeomColumn.setChecked(False)
        self.chkSpatialIndex.setChecked(False)
        self.editGeomSrid.clear()

        self.cboGeomType.setEnabled(False)
        self.editGeomColumn.setEnabled(False)
        self.spinGeomDim.setEnabled(False)
        self.editGeomSrid.setEnabled(False)
        self.chkSpatialIndex.setEnabled(False)

        QMessageBox.information(
            self, self.tr("DB Manager"), self.tr("Table created successfully.")
        )



"""
***************************************************************************
    DatasetIdentify.py
    ------------------
    Date                 : December 2025
    Copyright            : (C) 2025 by Even Rouault
    Email                : even dot rouault at spatialys dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Even Rouault"
__date__ = "December 2025"
__copyright__ = "(C) 2025, Even Rouault"

import os

from osgeo import gdal
from qgis.core import (
    Qgis,
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterFile,
    QgsProcessingParameterVectorDestination,
)

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class DatasetIdentify(GdalAlgorithm):
    INPUT = "INPUT"
    RECURSIVE = "RECURSIVE"
    DETAILS = "DETAILS"
    OUTPUT = "OUTPUT"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterFile(
                self.INPUT,
                self.tr("Input folder"),
                Qgis.ProcessingFileParameterBehavior.Folder,
            )
        )

        self.addParameter(
            QgsProcessingParameterBoolean(
                self.RECURSIVE,
                self.tr("Perform recursive exploration of the input folder"),
                defaultValue=True,
            )
        )

        detailsParam = QgsProcessingParameterBoolean(
            self.DETAILS,
            self.tr("Add details about identified datasets in the output"),
            defaultValue=True,
        )
        detailsParam.setHelp(
            self.tr(
                "Reports if the file is a Cloud Optimized GeoTIFF (COG), the list of files that compose the dataset, if it has georeferencing (has_geotransform and has_crs fields) and overviews"
            )
        )
        self.addParameter(detailsParam)

        self.addParameter(
            QgsProcessingParameterVectorDestination(
                self.OUTPUT,
                self.tr("Output file"),
                QgsProcessing.SourceType.TypeVector,
            )
        )

    def name(self):
        return "dataset_identify"

    def displayName(self):
        return self.tr("Dataset identification")

    def group(self):
        return self.tr("Miscellaneous")

    def groupId(self):
        return "miscellaneous"

    def commandName(self):
        return "gdal dataset identify"

    def getConsoleCommands(self, parameters, context, feedback, executing=True):
        arguments = []
        if self.parameterAsBoolean(parameters, self.RECURSIVE, context):
            arguments.append("--recursive")
        if self.parameterAsBoolean(parameters, self.DETAILS, context):
            arguments.append("--detailed")

        inDirectory = self.parameterAsFile(parameters, self.INPUT, context)
        if not os.path.exists(inDirectory):
            raise QgsProcessingException(
                self.tr("Directory {} does not exist").format(inDirectory)
            )
        arguments.append("--input")
        arguments.append(inDirectory)

        outFile = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)
        self.setOutputValue(self.OUTPUT, outFile)
        arguments.append("--output")
        arguments.append(outFile)

        return [self.commandName(), GdalUtils.escapeAndJoin(arguments)]

    def processAlgorithm(self, parameters, context, feedback):
        if int(gdal.VersionInfo()) < 3130000:
            raise QgsProcessingException(
                self.tr("GDAL 3.13 or later is needed to run this algorithm")
            )
        return super().processAlgorithm(parameters, context, feedback)

    def shortHelpString(self):
        return self.tr(
            "This algorithm reports the name of GDAL drivers that can open "
            "files contained in a folder, with optional additional details, "
            "and write the result into an output vector layer."
        )

    def shortDescription(self):
        return self.tr(
            "Reports the name of GDAL drivers that can open files contained in a folder, "
            "with optional details."
        )



"""
***************************************************************************
    v_proj.py
    ---------
    Date                 : November 2017
    Copyright            : (C) 2017 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "November 2017"
__copyright__ = "(C) 2017, Médéric Ribreux"

from qgis.core import QgsProcessingParameterString

from grassprovider.grass_utils import GrassUtils


def processInputs(alg, parameters, context, feedback):
    # Grab the projection from the input vector layer
    layer = alg.parameterAsLayer(parameters, "input", context)
    alg.setSessionProjectionFromLayer(layer, context)
    layerCrs = layer.crs().toProj()

    # Creates a new location with this Crs
    wkt_file_name = GrassUtils.exportCrsWktToFile(layer.crs(), context)
    newLocation = f"newProj{alg.uniqueSuffix}"
    alg.commands.append(f'g.proj wkt="{wkt_file_name}" location={newLocation}')

    # Go to the newly created location
    alg.commands.append(f"g.mapset mapset=PERMANENT location={newLocation}")

    # Import the layer
    alg.loadVectorLayerFromParameter("input", parameters, context, feedback, False)

    # Go back to default location
    alg.commands.append("g.mapset mapset=PERMANENT location=temp_location")

    # Grab the projected Crs
    crs = alg.parameterAsCrs(parameters, "crs", context)
    wkt_file_name = GrassUtils.exportCrsWktToFile(crs, context)
    alg.commands.append(f'g.proj -c wkt="{wkt_file_name}"')

    # Remove crs parameter
    alg.removeParameter("crs")

    # Add the location parameter with proper value
    location = QgsProcessingParameterString(
        "location", "new location", f"newProj{alg.uniqueSuffix}"
    )
    alg.addParameter(location)



"""
***************************************************************************
    ModelerTest
    ---------------------
    Date                 : November 2016
    Copyright            : (C) 2016 by Nyall Dawson
    Email                : nyall dot dawson at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************8
"""

__author__ = "Nyall Dawson"
__date__ = "November 2016"
__copyright__ = "(C) 2016, Nyall Dawson"

import unittest

from qgis.core import (
    QgsProcessingModelAlgorithm,
    QgsProcessingModelParameter,
    QgsProcessingParameterDistance,
    QgsProcessingParameterField,
    QgsProcessingParameterFile,
    QgsProcessingParameterNumber,
    QgsProcessingParameterString,
)
from qgis.testing import QgisTestCase, start_app

from processing.modeler.ModelerParametersDialog import ModelerParametersDialog

start_app()


class ModelerTest(QgisTestCase):
    def testModelerParametersDialogAvailableValuesOfType(self):
        # test getAvailableValuesOfType from ModelerParametersDialog

        m = QgsProcessingModelAlgorithm()

        string_param_1 = QgsProcessingModelParameter("string")
        m.addModelParameter(QgsProcessingParameterString("string"), string_param_1)

        string_param_2 = QgsProcessingModelParameter("string2")
        m.addModelParameter(QgsProcessingParameterString("string2"), string_param_2)

        num_param = QgsProcessingModelParameter("number")
        m.addModelParameter(QgsProcessingParameterNumber("number"), num_param)

        table_field_param = QgsProcessingModelParameter("field")
        m.addModelParameter(QgsProcessingParameterField("field"), table_field_param)

        file_param = QgsProcessingModelParameter("file")
        m.addModelParameter(QgsProcessingParameterFile("file"), file_param)

        dlg = ModelerParametersDialog(m, m)
        # test single types
        self.assertEqual(
            {
                p.parameterName()
                for p in dlg.getAvailableValuesOfType(QgsProcessingParameterNumber)
            },
            {"number"},
        )
        self.assertEqual(
            {
                p.parameterName()
                for p in dlg.getAvailableValuesOfType(QgsProcessingParameterField)
            },
            {"field"},
        )
        self.assertEqual(
            {
                p.parameterName()
                for p in dlg.getAvailableValuesOfType(QgsProcessingParameterFile)
            },
            {"file"},
        )

        # test multiple types
        self.assertEqual(
            {
                p.parameterName()
                for p in dlg.getAvailableValuesOfType(
                    [
                        QgsProcessingParameterString,
                        QgsProcessingParameterNumber,
                        QgsProcessingParameterFile,
                    ]
                )
            },
            {"string", "string2", "number", "file"},
        )


if __name__ == "__main__":
    unittest.main()



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : May 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

import os
from importlib import import_module

current_dir = os.path.dirname(__file__)


def load(db, mainwindow):
    for name in os.listdir(current_dir):
        if not os.path.isdir(os.path.join(current_dir, name)):
            continue
        if name in ("__pycache__"):
            continue
        try:
            plugin_module = import_module(".".join((__package__, name)))
        except ImportError:
            continue
        plugin_module.load(db, mainwindow)



"""QGIS Unit tests for Grass Algorithm with postgreraster provider

.. note:: This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
"""

__author__ = "Jan Caha"
__date__ = "01/08/2025"
__copyright__ = "Copyright 2025, The QGIS Project"


import os
import tempfile
import unittest
from shutil import rmtree

from grassprovider.grass_provider import GrassProvider
from grassprovider.grass_utils import GrassUtils
from qgis import processing
from qgis.core import (
    QgsApplication,
    QgsAuthMethodConfig,
    QgsDataSourceUri,
    QgsProcessingContext,
    QgsProcessingFeedback,
    QgsRasterLayer,
)
from qgis.testing import QgisTestCase, start_app

QGIS_AUTH_DB_DIR_PATH = tempfile.mkdtemp()
os.environ["QGIS_AUTH_DB_DIR_PATH"] = QGIS_AUTH_DB_DIR_PATH

start_app()


class TestProcessingGrassAlgsPostgreRasterProvider(QgisTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.provider = GrassProvider()
        QgsApplication.processingRegistry().addProvider(cls.provider)
        cls.cleanup_paths = []

        assert GrassUtils.installedVersion()

    @classmethod
    def tearDownClass(cls):
        """Run after all tests"""
        rmtree(QGIS_AUTH_DB_DIR_PATH)
        del os.environ["QGIS_AUTH_DB_DIR_PATH"]
        super().tearDownClass()

    def test_algorithm_r_buffer(self):
        """
        Test grass algorithm r.buffer with postgreraster provider
        """

        context = QgsProcessingContext()
        feedback = QgsProcessingFeedback()

        rl = QgsRasterLayer(
            "dbname='mydb' host=localhost port=5432 user='asdf' password='42'"
            " sslmode=disable table=some_table schema=some_schema column=rast sql=pk = 2",
            "pg_layer",
            "postgresraster",
        )

        alg = QgsApplication.processingRegistry().algorithmById("grass:r.buffer")

        self.assertTrue(alg)

        params = {
            "input": rl,
            "distances": "100,200,500",
            "units": 0,
            "-z": False,
            "output": "TEMPORARY_OUTPUT",
            "GRASS_REGION_PARAMETER": None,
            "GRASS_REGION_CELLSIZE_PARAMETER": 0,
            "GRASS_RASTER_FORMAT_OPT": "",
            "GRASS_RASTER_FORMAT_META": "",
        }

        can_run, _ = alg.checkParameterValues(parameters=params, context=context)

        self.assertTrue(can_run)

        res = alg.run(params, context, feedback)

        # should be tuple
        self.assertTrue(res)
        # should be true if algorithm run successfully
        self.assertTrue(res[1])
        # should be dict with output keys
        self.assertTrue(isinstance(res[0], dict))
        # should be path to result file
        self.assertTrue(isinstance(res[0]["output"], str))

    def test_algorithm_r_info(self):
        """
        Test grass algorithm r.info with postgreraster provider
        """

        context = QgsProcessingContext()
        feedback = QgsProcessingFeedback()

        rl = QgsRasterLayer(
            "dbname='mydb' host=localhost port=5432 user='asdf' password='42'"
            " sslmode=disable table=some_table schema=some_schema column=rast sql=pk = 2",
            "pg_layer",
            "postgresraster",
        )

        alg = QgsApplication.processingRegistry().algorithmById("grass:r.info")

        params = {
            "map": rl,
            "-r": False,
            "-g": False,
            "-h": False,
            "-e": False,
            "html": "./report.html",
            "GRASS_REGION_PARAMETER": None,
            "GRASS_REGION_CELLSIZE_PARAMETER": 0,
        }

        can_run, _ = alg.checkParameterValues(parameters=params, context=context)

        self.assertTrue(can_run)

        res = alg.run(params, context, feedback)

        # should be tuple
        self.assertTrue(res)
        # should be true if algorithm run successfully
        self.assertTrue(res[1])
        # should be dict with output keys
        self.assertTrue(isinstance(res[0], dict))
        # should be path to result file
        self.assertTrue(isinstance(res[0]["html"], str))


if __name__ == "__main__":
    unittest.main()
